###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import six

if six.PY2:
    from pkgutil import extend_path

    __path__ = extend_path(__path__, __name__)


# Define Version
if six.PY3:
    from pkg_resources import get_distribution, DistributionNotFound

    try:
        __version__ = get_distribution(__name__).version
        version = __version__
    except DistributionNotFound:
        # package is not installed
        version = "Unknown"
else:
    majorVersion = 5
    minorVersion = 3
    patchLevel = 1
    preVersion = 0

    version = "v%sr%s" % (majorVersion, minorVersion)
    buildVersion = "v%dr%d" % (majorVersion, minorVersion)
    if patchLevel:
        version = "%sp%s" % (version, patchLevel)
        buildVersion = "%s build %s" % (buildVersion, patchLevel)
    if preVersion:
        version = "%s-pre%s" % (version, preVersion)
        buildVersion = "%s pre %s" % (buildVersion, preVersion)


def extension_metadata():
    import importlib.resources  # pylint: disable=import-error,no-name-in-module

    return {
        "priority": 60,
        "web_resources": {
            "static": [importlib.resources.files("LHCbWebDIRAC") / "WebApp" / "static"],  # pylint: disable=no-member
        },
    }



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import json

import six

from DIRAC import S_OK, S_ERROR, gConfig, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client import PathFinder
from WebAppDIRAC.Lib.WebHandler import WebHandler, asyncGen, WErr

from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

# TODO: Replace with
# from LHCbDIRAC.ConfigurationSystem.Client.Helpers.Resources import DEFAULT_CACHEPATH
DEFAULT_CACHEPATH = "/cvmfs/lhcb.cern.ch/lib/var/lib/softmetadata/project-platforms.json"

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError


def getSoftVersions(name, web_names="", add=""):
    """Return list of all downloadable versions
    of specified software package.
    web_name is default to name_name
    add specify verbatim version to add to the list
    Note: It was in ProductionLib
    """

    if web_names == "":
        web_names = name

    try:
        with open(DEFAULT_CACHEPATH, "rt") as fp:
            projectMetadata = json.load(fp)
    except (FileNotFoundError, JSONDecodeError):
        gLogger.exception("Failed to load projects from", DEFAULT_CACHEPATH)
        return S_ERROR("Failed to load projects from %s" % DEFAULT_CACHEPATH)

    vers = set()
    for web_name in web_names.split(","):
        platforms = projectMetadata.get(web_name.upper())
        if not platforms:
            return S_ERROR("%s is not in the list of known platforms" % web_name)
        vers |= set(platforms)

    # Remove duplicates
    allversions = tuple(vers)
    if add:
        allversions += tuple(add.split(","))

    # Sort the result, newer version will be first
    allversions = list(sorted(allversions, reverse=True))
    gLogger.debug("Application versions", allversions)
    return S_OK(allversions)


class LHCbStepManagerHandler(WebHandler):

    AUTH_PROPS = "authenticated"

    def __getArgument(self, argName, argDefValue):
        return self.get_argument(argName, None) or argDefValue

    def __getJsonArgument(self, argName, argDefValue):
        x = self.get_argument(argName, None)
        return json.loads(x) if x else argDefValue

    def __getApplications(self):
        applications = []
        csS = PathFinder.getServiceSection("ProductionManagement/ProductionRequest")
        if csS:
            applications = gConfig.getValue("%s/Applications" % csS, "")
        if not applications:
            applications = ["Gauss", "Boole", "Brunel", "DaVinci", "Moore", "LHCb"]
        else:
            applications = applications.split(", ")
        return applications

    def web_getApplications(self):
        applications = self.__getApplications()

        self.write({"success": "true", "total": len(applications), "result": [{"v": x} for x in applications]})

    @asyncGen
    def web_getAppVersions(self):
        app = self.__getArgument("app", "")
        if not app:
            self.finish({"success": "true", "result": [], "total": 0})
            return

        result = yield self.threadTask(getSoftVersions, app)
        if not result["OK"]:
            self.finish({"success": "false", "result": [], "total": 0, "error": result["Message"]})
            return

        versions = result["Value"]
        self.finish({"success": "true", "total": len(versions), "result": [{"v": x} for x in versions]})

    def web_getAppOptionsFormats(self):
        app = self.__getArgument("app", "")
        if not app:
            self.write({"success": "true", "result": [], "total": 0})
            return

        formats = Operations().getValue("Productions/StepOptionsFormat/%s" % app)
        formats = formats.split(", ") if formats is not None else []
        self.finish({"success": "true", "total": len(formats), "result": [{"v": x} for x in formats]})

    def web_getSelectionData(self):
        callback = {
            "ApplicationName": [[x] for x in self.__getApplications()],
            "Visible": [["Yes"], ["No"]],
            "Usable": [["Yes"], ["Not ready"], ["Obsolete"]],
        }
        self.write(callback)

    def __getFilter(self):
        visibleMap = {"Yes": "Y", "No": "N"}
        sfilter = {}

        # Selector
        try:

            for selFieldName in [
                "ApplicationName",
                "ApplicationVersion",
                "Visible",
                "Usable",
                "ProcessingPass",
                "StartDate",
                "StartDate",
                "InputFileTypes",
                "OutputFileTypes",
                "Equal",
                "StepId",
            ]:
                field = self.__getJsonArgument(selFieldName, [])
                if field:
                    if selFieldName == "Equal":
                        sfilter[selFieldName] = str(field[-1])
                    elif selFieldName == "StartDate":
                        sfilter[selFieldName] = str(field[0])
                    elif selFieldName == "Visible":
                        sfilter[selFieldName] = [visibleMap.get(str(x), "Y") for x in field]
                    else:
                        sfilter[selFieldName] = [str(x) for x in field]
                        gLogger.info("%s = %s" % (selFieldName, str(sfilter[selFieldName])))
        except Exception as e:
            gLogger.info("__getFilter: Wrong selection: %s" % str(e))

        # Grid ordering and sorting
        sort = {"Items": "StepId", "Order": "Desc"}
        try:
            start = int(self.__getArgument("start", 0))
            limit = int(self.__getArgument("limit", 0))
            sortlist = self.__getJsonArgument("sort", [])
            if sortlist:
                sort = {"Items": str(sortlist[-1]["property"]), "Order": str(sortlist[-1]["direction"])}
        except Exception:  # fallback to defaults instead of error
            start = 0
            limit = 25
        if limit > 0:
            sfilter["StartItem"] = start
            sfilter["MaxItem"] = start + limit
        sfilter["Sort"] = sort
        gLogger.info(str(sfilter))
        return sfilter

    def __runtimeProjectsConvert(self, step):
        projects = step.get("RuntimeProjects", [])
        if projects:
            fieldNames = projects["ParameterNames"]
            projects = [dict(zip(fieldNames, y)) for y in projects["Records"]]
        step["RuntimeProjects"] = projects
        step["textRuntimeProjects"] = ",".join(
            ["%s(%s)" % (x.get("StepName", ""), x.get("StepId", "")) for x in projects]
        )

    def __nullConvert(self, oneDict):
        for x in oneDict:
            if oneDict[x] is None:
                oneDict[x] = ""

    @asyncGen
    def web_getSteps(self):
        sfilter = self.__getFilter()

        result = yield self.threadTask(BookkeepingClient().getAvailableSteps, sfilter)

        if not result["OK"]:
            self.finish({"success": "false", "result": [], "total": 0, "error": result["Message"]})
            return

        result = result["Value"]

        if "TotalRecords" in result:
            if not (result["TotalRecords"] > 0):
                raise WErr(500, "There were no data matching your selection")
        else:
            raise WErr(500, "There were no data matching your selection")

        fields = result["ParameterNames"]
        steps = [dict(zip(fields, x)) for x in result["Records"]]
        for step in steps:
            self.__runtimeProjectsConvert(step)

        self.finish({"success": "true", "result": steps, "total": result["TotalRecords"], "date": None})

    @asyncGen
    def web_getStep(self):
        bk = BookkeepingClient()

        StepId = 0
        try:
            StepId = int(self.__getArgument("StepId", 0))
        except Exception:
            pass  # ignore errors

        # Get Step Body
        result = yield self.threadTask(bk.getAvailableSteps, {"StepId": StepId})
        if not result["OK"]:
            self.finish({"success": "false", "result": [], "error": result["Message"]})
            return
        result = result["Value"]
        fields = result["ParameterNames"]
        steps = [dict(zip(fields, x)) for x in result["Records"]]
        if len(steps) != 1:
            self.finish({"success": "false", "result": [], "error": "Requested Step is not found"})
            return
        step = steps[0]

        # Get Input Files for the step
        result = yield self.threadTask(bk.getStepInputFiles, StepId)
        if not result["OK"]:
            self.finish({"success": "false", "result": [], "error": result["Message"]})
            return
        result = result["Value"]
        fields = ["FileType"]  # TODO: replace with result['ParameterNames']
        ift = [dict(zip(fields, x)) for x in result["Records"]]

        # Get Output Files for the step
        result = yield self.threadTask(bk.getStepOutputFiles, StepId)
        if not result["OK"]:
            self.finish({"success": "false", "result": [], "error": result["Message"]})
            return
        result = result["Value"]
        fields = ["FileType"]  # TODO: replace with result['ParameterNames']
        oft = [dict(zip(fields, x)) for x in result["Records"]]

        # Put everything together
        step["InputFileTypes"] = ift
        step["OutputFileTypes"] = oft
        try:
            step["textInputFileTypes"] = ",".join(["%s" % x["FileType"] for x in ift])
            step["textOutputFileTypes"] = ",".join(["%s" % x["FileType"] for x in oft])
        except Exception as e:
            self.finish({"success": "false", "result": [], "error": "Can not convert File Types: %s" % str(e)})
        self.__runtimeProjectsConvert(step)
        self.__nullConvert(step)

        self.finish({"success": "true", "result": step})

    @asyncGen
    def web_getRuntimeProjects(self):

        result = yield self.threadTask(BookkeepingClient().getAvailableSteps, {})
        if not result["OK"]:
            self.finish({"success": "false", "result": [], "error": result["Message"]})
            return

        fields = result["Value"]["ParameterNames"]
        rows = [dict(zip(fields, x)) for x in result["Value"]["Records"]]
        rows = [
            {"id": r["StepId"], "text": "%s(%s)" % (r["StepName"], r["StepId"])} for r in rows if r["Usable"] == "Yes"
        ]
        rows.sort(key=lambda x: x["id"], reverse=True)
        self.finish({"success": "true", "result": rows, "total": len(rows)})

    @asyncGen
    def web_getBKTags(self):
        tag = self.__getArgument("tag", "")
        if not tag:
            self.finish({"success": "true", "result": [], "total": 0})
            return

        result = yield self.threadTask(BookkeepingClient().getAvailableTagsFromSteps)
        if not result["OK"]:
            self.finish({"success": "false", "result": [], "total": 0, "error": result["Message"]})
            return

        rows = [{"v": x[1]} for x in result["Value"]["Records"] if x[0] == tag]
        rows.sort(key=lambda x: x["v"])

        if "ONLINE" not in [tag.values()[0] for tag in rows]:
            rows.append({"v": "ONLINE"})
        if "fromPreviousStep" not in [tag.values()[0] for tag in rows]:
            rows.append({"v": "fromPreviousStep"})
        self.finish({"success": "true", "total": len(rows), "result": rows})

    @asyncGen
    def web_getFileTypes(self):
        result = yield self.threadTask(BookkeepingClient().getAvailableFileTypes)
        if not result["OK"]:
            self.finish({"success": "false", "result": [], "total": 0, "error": result["Message"]})
            return
        rows = [dict(zip(["Name", "Description"], x)) for x in result["Value"]["Records"]]
        rows.sort(key=lambda x: x["Name"])
        self.finish({"success": "true", "total": len(rows), "result": rows})

    @asyncGen
    def web_addFileType(self):
        name = self.__getArgument("Name", "")
        description = self.__getArgument("Description", "")
        if not name or not description:
            self.finish({"success": "false", "result": [], "error": "File type specification is incomplete"})
            return
        result = yield self.threadTask(BookkeepingClient().insertFileTypes, str(name), str(description), "ROOT")
        if not result["OK"]:
            self.finish({"success": "false", "result": [], "error": result["Message"]})
        else:
            self.finish({"success": "true", "result": []})

    __stepOrdinaryFields = [
        "StepId",
        "StepName",
        "ApplicationName",
        "ApplicationVersion",
        "SystemConfig",
        "mcTCK",
        "OptionFiles",
        "OptionsFormat",
        "DDDB",
        "CONDDB",
        "DQTag",
        "ExtraPackages",
        "Visible",
        "Usable",
        "ProcessingPass",
        "isMulticore",
    ]

    def __decodeFileTypes(self, s):
        r = []
        try:
            ftl = json.loads(s)
            for x in ftl:
                d = {}
                for y in x:
                    d[str(y)] = str(x[y])
                r.append(d)
        except Exception as e:
            gLogger.error("Cound not covert 's': %s" % str(e))
        return r

    @asyncGen
    def web_saveStep(self):
        params = {}
        for name in self.request.arguments:
            value = self.get_argument(name)
            if name in self.__stepOrdinaryFields:
                if six.PY2 and isinstance(value, unicode):  # pylint: disable=undefined-variable
                    value = str(value)
                params[name] = value
            elif name in ["InputFileTypes", "OutputFileTypes"]:
                params[name] = self.__decodeFileTypes(value)
            elif name in ["RuntimeProjectStepId"]:
                if value:
                    params["RuntimeProjects"] = [{"StepId": int(value)}]
        if "RuntimeProjects" not in params:
            params["RuntimeProjects"] = []

        if "StepId" not in params:
            params["StepId"] = "0"

        bk = BookkeepingClient()
        if params["StepId"] is "0":
            for name in ["StepId", "InputFileTypes", "OutputFileTypes", "RuntimeProjects"]:
                if name in params and not params[name]:
                    del params[name]
            params = {"Step": params}
            for name in ["InputFileTypes", "OutputFileTypes"]:
                if name in params["Step"]:
                    params[name] = params["Step"][name]
                    del params["Step"][name]
            result = yield self.threadTask(bk.insertStep, params)
        else:
            result = yield self.threadTask(bk.updateStep, params)
        if not result["OK"]:
            self.finish({"success": "false", "result": [], "error": result["Message"]})
        else:
            self.finish({"success": "true", "result": []})

    @asyncGen
    def web_deleteStep(self):
        StepId = 0
        try:
            StepId = int(self.__getArgument("StepId", 0))
        except Exception:
            pass
        if not StepId:
            self.finish({"success": "false", "result": [], "error": "StepId is not correctly specified"})
            return
        result = yield self.threadTask(BookkeepingClient().deleteStep, StepId)
        if not result["OK"]:
            self.finish({"success": "false", "result": [], "error": result["Message"]})
        else:
            self.finish({"success": "true", "result": []})



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import json

from WebAppDIRAC.WebApp.handler.JobMonitorHandler import JobMonitorHandler
from WebAppDIRAC.Lib.SessionData import SessionData


class LHCbJobMonitorHandler(JobMonitorHandler):

    AUTH_PROPS = "authenticated"

    def index(self):
        pass

    def web_standalone(self):
        self.render("JobMonitor/standalone.tpl", config_data=json.dumps(SessionData(None, None).getData()))

    def _request(self):
        req = super(LHCbJobMonitorHandler, self)._request()
        runs = list(json.loads(self.get_argument("RunNumbers", "[]")))
        if runs:
            req["runNumber"] = runs
        return req



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import json

import six

from DIRAC import gLogger
from DIRAC.Core.Utilities import Time
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from WebAppDIRAC.Lib.WebHandler import WebHandler, asyncGen


class BookkeepingSimDescriptionHandler(WebHandler):

    AUTH_PROPS = "authenticated"

    def index(self):
        pass

    @asyncGen
    def web_getSelectionData(self):
        data = {"Visible": ["Y", "N"]}
        self.finish(data)

    @asyncGen
    def web_getData(self):

        filter = {}
        limit = 0
        dir = "DESC"
        sort = "SimId"
        data = None

        try:
            for f in ["SimId", "SimDescription", "Visible"]:
                jsvalue = json.loads(self.get_argument(f, "[]"))
                v = str(jsvalue[0]) if len(jsvalue) > 0 else ""
                if v != "":
                    filter[f] = v

            start = int(self.get_argument("start", "0"))
            limit = int(self.get_argument("limit", "0"))
            if "sort" in self.request.arguments:
                data = json.loads(self.get_argument("sort"))
                sort = str(data[-1]["property"])
                dir = str(data[-1]["direction"])

        except Exception as e:
            data = {"success": "false", "error": str(e)}

        if limit > 0:
            filter["StartItem"] = start
            filter["MaxItem"] = start + limit

        if dir == "ASC":
            filter["Sort"] = {"Items": sort, "Order": "Asc"}
        else:
            filter["Sort"] = {"Items": sort, "Order": "Desc"}

        gLogger.debug("getSimulatuionConditions", str(filter))

        retVal = yield self.threadTask(BookkeepingClient().getSimulationConditions, filter)

        if not retVal["OK"]:
            data = {"success": "false", "error": retVal["Message"]}
        else:
            timestamp = Time.dateTime().strftime("%Y-%m-%d %H:%M [UTC]")

            fields = retVal["Value"]["ParameterNames"]
            totalRecords = retVal["Value"]["TotalRecords"]
            rows = [dict(zip(fields, i)) for i in retVal["Value"]["Records"]]
            data = {"success": "true", "result": rows, "total": totalRecords, "date": timestamp}
        self.finish(data)

    @asyncGen
    def web_editSimulation(self):
        try:
            simId = int(self.get_argument("SimId"))
        except ValueError:
            self.finish("SimId is not a number")

        result = yield self.threadTask(BookkeepingClient().getSimulationConditions, {"SimId": simId})
        if not result["OK"]:
            self.finish({"success": "false", "error": result["Message"]})
        fields = result["Value"]["ParameterNames"]
        row = [dict(zip(fields, x)) for x in result["Value"]["Records"]][0]
        self.finish({"success": "true", "data": row})

    @asyncGen
    def web_simulationinsert(self):
        simdict = {k: self.get_argument(k) for k in self.request.arguments}
        if "SimId" in simdict:
            del simdict["SimId"]
        for i, value in simdict.items():
            if six.PY2 and isinstance(value, unicode):  # pylint: disable=undefined-variable
                simdict[i] = str(simdict[i])
            else:
                simdict[i] = simdict[i]
        gLogger.debug("Insert:", str(simdict))

        retVal = yield self.threadTask(BookkeepingClient().insertSimConditions, simdict)
        result = None
        if retVal["OK"]:
            result = {"success": "true", "result": "It is registered to the database!"}
        else:
            result = {"success": "false", "error": retVal["Message"]}

        self.finish(result)

    @asyncGen
    def web_simulationupdate(self):
        simdict = {k: self.get_argument(k) for k in self.request.arguments}
        for i, value in simdict.items():
            if six.PY2 and isinstance(value, unicode):  # pylint: disable=undefined-variable
                simdict[i] = str(simdict[i])
            else:
                simdict[i] = simdict[i]
        gLogger.debug("Insert:", str(simdict))
        retVal = yield self.threadTask(BookkeepingClient().updateSimulationConditions, simdict)
        result = None
        if retVal["OK"]:
            result = {"success": "true", "result": "It has successfully updated!"}
        else:
            result = {"success": "false", "error": retVal["Message"]}

        self.finish(result)

    @asyncGen
    def web_simulationdelete(self):
        try:
            simId = int(self.get_argument("SimId"))
        except ValueError as e:
            self.finish({"success": "false", "error": str(e)})
        gLogger.debug("SimId:", id)
        retVal = yield self.threadTask(BookkeepingClient().deleteSimulationConditions, simId)
        result = None
        if retVal["OK"]:
            result = {"success": "true", "result": "It has successfully deleted!"}
        else:
            result = {"success": "false", "error": retVal["Message"]}

        self.finish(result)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" Web Handler for ProductionRequest service
"""
import json
import datetime
import tornado
from six import text_type
import six

from WebAppDIRAC.Lib.WebHandler import WebHandler, WErr, asyncGen

from DIRAC import S_ERROR, S_OK, gConfig, gLogger
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.Core.Utilities import DictCache
from DIRAC.Core.Utilities import Time

from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
from LHCbWebDIRAC.WebApp.handler.ProductionLib import SelectAndSort, PrTpl
from LHCbDIRAC.BookkeepingSystem.Client.LHCB_BKKDBClient import LHCB_BKKDBClient


class ProductionRequestManagerHandler(WebHandler):

    AUTH_PROPS = "authenticated"

    __dataCache = DictCache.DictCache()

    serviceFields = [
        "RequestID",
        "HasSubrequest",
        "ParentID",
        "MasterID",
        "RequestName",
        "RequestType",
        "RequestState",
        "RequestPriority",
        "RequestAuthor",
        "RequestPDG",
        "RequestWG",
        "SimCondition",
        "SimCondID",
        "ProPath",
        "ProID",
        "EventType",
        "NumberOfEvents",
        "Comments",
        "Description",
        "Inform",
        "IsModel",
        "StartingDate",
        "FinalizationDate",
        "FastSimulationType",
        "RetentionRate",
        "bk",
        "bkTotal",
        "rqTotal",
        "crTime",
        "upTime",
    ]
    localFields = [
        "ID",
        "_is_leaf",
        "_parent",
        "_master",
        "reqName",
        "reqType",
        "reqState",
        "reqPrio",
        "reqAuthor",
        "reqPDG",
        "reqWG",
        "simDesc",
        "simCondID",
        "pDsc",
        "pID",
        "eventType",
        "eventNumber",
        "reqComment",
        "reqDesc",
        "reqInform",
        "IsModel",
        "StartingDate",
        "FinalizationDate",
        "FastSimulationType",
        "RetentionRate",
        "eventBK",
        "eventBKTotal",
        "EventNumberTotal",
        "creationTime",
        "lastUpdateTime",
    ]

    simcondFields = [
        "Generator",
        "MagneticField",
        "BeamEnergy",
        "Luminosity",
        "DetectorCond",
        "BeamCond",
        "configName",
        "configVersion",
        "condType",
        "inProPass",
        "inFileType",
        "inProductionID",
        "inDataQualityFlag",
        "G4settings",
        "inTCKs",
    ]

    proStepFields = [
        "Step",
        "Name",
        "Pass",
        "App",
        "Ver",
        "Opt",
        "OptF",
        "DDDb",
        "CDb",
        "DQT",
        "EP",
        "Vis",
        "Use",
        "IFT",
        "OFT",
        "Html",
    ]

    extraFields = ["mcConfigVersion"]

    bkSimCondFields = [
        "simCondID",
        "simDesc",
        "BeamCond",
        "BeamEnergy",
        "Generator",
        "MagneticField",
        "DetectorCond",
        "Luminosity",
        "G4settings",
    ]

    def __unescape(self, d):
        """Unescape HTML code in parameters"""
        for x in d:
            d[x] = text_type(d[x][-1])
            s = d[x].replace("&amp;", "&")
            while s != d[x]:
                d[x] = s
                s = d[x].replace("&amp;", "&")
            d[x] = d[x].replace("&lt;", "<")
            d[x] = d[x].replace("&gt;", ">")

    def __fromlocal(self, req):
        result = {}
        for x, y in zip(self.localFields[2:-5], self.serviceFields[2:-5]):
            if x == "eventType" and "eventType" in req:
                result[y] = req[x]
            elif x in req and req[x]:
                result[y] = req[x]

        SimCondDetail = {}
        for x in self.simcondFields:
            if x in req and str(req[x]) != "":
                SimCondDetail[x] = req[x]
        if SimCondDetail:
            result["SimCondDetail"] = json.dumps(SimCondDetail)

        ProDetail = {}
        for x in req:
            if str(x)[0] == "p" and str(x) != "progress":
                if str(req[x]) != "":
                    ProDetail[x] = req[x]
        if ProDetail:
            result["ProDetail"] = json.dumps(ProDetail)

        Extra = {}
        for x in self.extraFields:
            if x in req and str(req[x]) != "":
                Extra[x] = req[x]
        if Extra:
            result["Extra"] = json.dumps(Extra)

        return result

    def __request_types(self):
        types = ""
        csS = PathFinder.getServiceSection("ProductionManagement/ProductionRequest")
        if csS:
            types = gConfig.getValue("%s/RequestTypes" % csS, "")
        if not types:
            return ["Simulation", "Reconstruction", "Stripping"]
        return types.split(", ")

    def __getArgument(self, argName, argDefValue):
        return self.get_argument(argName, None) or argDefValue

    def __getJsonArgument(self, argName, argDefValue):
        x = self.get_argument(argName, None)
        return json.loads(x) if x else argDefValue

    @staticmethod
    def __bkProductionProgress(prId):
        result = BookkeepingClient().getProductionInformations(prId)
        if not result["OK"]:
            return result
        info = result["Value"]
        if "Number of events" not in info:
            return S_OK(0)
        allevents = info["Number of events"]
        if not allevents:
            return S_OK(0)
        if len(allevents) > 1:
            return S_ERROR("More than one output file type. Unsupported.")
        return S_OK(allevents[0][1])

    def __2local(self, req):
        result = {}

        for x, y in zip(self.localFields, self.serviceFields):
            if isinstance(req[y], (datetime.date, datetime.datetime)):
                result[x] = str(req[y])
            else:
                result[x] = req[y]
        result["_is_leaf"] = not result["_is_leaf"]
        if req["bkTotal"] is not None and req["rqTotal"] is not None and req["rqTotal"]:
            result["progress"] = int(req["bkTotal"]) * 100 / int(req["rqTotal"])
        else:
            result["progress"] = None
        if req["SimCondDetail"]:
            try:
                result.update(json.loads(req["SimCondDetail"]))
            except Exception:
                gLogger.error("Bad requests: simulation/data taking conditions is not defined", req["RequestID"])

        if req["ProDetail"]:
            try:
                result.update(json.loads(req["ProDetail"]))
            except Exception:
                gLogger.error("Bad requests: processing pass is not defined", req["RequestID"])

        if req["Extra"]:
            result.update(json.loads(req["Extra"]))
        for x, value in result.items():
            if x != "_parent" and value is None:
                result[x] = ""
        return result

    def __parseRequest(self):
        path = self.get_argument("fullpath", "")
        rType = self.get_argument("type", "") == "adv"
        tree = self.get_argument("tree", "Configuration")

        dataQuality = None
        if "dataQuality" in self.request.arguments:
            dataQuality = dict(json.loads(self.get_argument("dataQuality")))
            if not dataQuality:
                dataQuality = {"OK": True}

        self.numberOfJobs = int(self.get_argument("limit", "25"))
        self.pageNumber = int(self.get_argument("start", "0"))
        return path, rType, tree, dataQuality

    def __condFromPath(self, path, bkcli, item):
        runCondTr = {
            "DaqperiodId": "simCondID",
            "Description": "simDesc",
            "BeamCondition": "BeamCond",
            "BeanEnergy": "BeamEnergy",
            "MagneticField": "MagneticField",
            "VELO": "detVELO",
            "IT": "detIT",
            "TT": "detTT",
            "OT": "detOT",
            "RICH1": "detRICH1",
            "RICH2": "detRICH2",
            "SPD_PRS": "detSPD_PRS",
            "ECAL": "detECAL",
            "HCAL": "detHCAL",
            "MUON": "detMUON",
            "L0": "detL0",
            "HLT": "detHLT",
            "VeloPosition": "VeloPosition",
        }
        simCondTr = {
            "BeamEnergy": "BeamEnergy",
            "Description": "simDesc",
            "Generator": "Generator",
            "Luminosity": "Luminosity",
            "G4settings": "G4settings",
            "MagneticField": "MagneticField",
            "DetectorCondition": "DetectorCond",
            "BeamCondition": "BeamCond",
            "SimId": "simCondID",
        }

        p = path.split("/")
        condp = "/".join(p[0:4])
        if not item:
            dirp = "/".join(p[0:3])
            allcond = bkcli.list(dirp)
            cond = {}
            for x in allcond:
                if x["fullpath"] == condp:
                    cond = x
                    break
            if not cond:
                return S_ERROR("Could not find %s" % condp)
        else:
            cond = item
        if cond["level"] != "Simulation Conditions/DataTaking":
            return S_ERROR("%s is not a condition" % condp)
        for x in ["level", "expandable", "name", "fullpath", "selection", "method"]:
            if x in cond:
                del cond[x]
        if "DaqperiodId" in cond:
            tr, rType = runCondTr, "Run"
        else:
            tr, rType = simCondTr, "Simulation"
        nm = [x for x in cond if x not in tr]
        if nm:
            return S_ERROR("Unmached run conditions: %s" % str(nm))
        v = dict([(tr[x], cond[x]) for x in cond])
        v["condType"] = rType
        if rType == "Run":
            v["DetectorCond"] = self.__runDetectorCond(v)
        return S_OK(v)

    def __runDetectorCond(self, cond):
        subd = [
            "detVELO",
            "detIT",
            "detTT",
            "detOT",
            "detRICH1",
            "detRICH2",
            "detSPD_PRS",
            "detECAL",
            "detHCAL",
            "detMUON",
            "detL0",
            "detHLT",
        ]
        incl = []
        not_incl = []
        unknown = []
        for x in subd:
            if x not in cond:
                unknown.append(x[3:])
            elif cond[x] == "INCLUDED":
                incl.append(x[3:])
            elif cond[x] == "NOT INCLUDED":
                not_incl.append(x[3:])
            else:
                unknown.append(x[3:])
        if len(incl) > len(not_incl):
            if not not_incl:
                s = "all"
                if unknown:
                    s += " except unknown %s" % ",".join(unknown)
            else:
                s = "without %s" % ",".join(not_incl)
                if unknown:
                    s += " and unknown %s" % ",".join(unknown)
        else:
            if incl:
                s = "only %s" % ",".join(incl)
                if unknown:
                    s += " and unknown %s" % ",".join(unknown)
            else:
                s = "no"
                if unknown:
                    s += " with unknown %s" % ",".join(unknown)
        return s

    @asyncGen
    def web_getSelectionData(self):
        callback = {}
        retVal = yield self.threadTask(ProductionRequestClient().getFilterOptions)
        if not retVal["OK"]:
            self.finish({"success": "false", "result": [], "total": 0, "error": retVal["Message"]})
            return

        callback["typeF"] = [[i] for i in retVal["Value"]["Type"]]
        callback["stateF"] = [[i] for i in retVal["Value"]["State"]]
        callback["authorF"] = [[i] for i in retVal["Value"]["Author"]]
        callback["wgF"] = [[i] for i in retVal["Value"]["WG"]]

        self.write(callback)

    @asyncGen
    def web_getWG(self):
        retVal = yield self.threadTask(ProductionRequestClient().getFilterOptions)
        if not retVal["OK"]:
            raise WErr.fromSERROR(retVal)
        rows = [{"text": i, "name": i} for i in retVal["Value"]["WG"]]
        self.finish({"OK": True, "result": rows, "total": len(rows)})

    @asyncGen
    def web_list(self):
        parent = self.get_argument("anode", 0)
        try:
            if parent != 0:
                parent = json.loads(parent)
                if isinstance(parent, list):
                    parent = parent[-1]
        except Exception as e:
            raise WErr(404, e)

        result = yield self.threadTask(self.getProductionRequestList, parent)
        self.finish(result)

    @asyncGen
    def web_history(self):
        requestID = self.get_argument("RequestID", "")
        try:
            requestID = int(json.loads(requestID))
        except Exception:
            self.finish({"success": "false", "result": [], "total": 0, "error": "Request ID is not a number"})

        result = yield self.threadTask(ProductionRequestClient().getRequestHistory, requestID)
        if not result["OK"]:
            self.finish({"success": "false", "result": [], "total": 0, "error": result["Message"]})
            return
        rows = result["Value"]["Rows"]
        result = []
        for row in rows:
            row["TimeStamp"] = str(row["TimeStamp"])
            result += [[row["TimeStamp"], row["RequestState"], row["RequestUser"]]]
        self.finish({"success": "true", "result": result, "total": len(result)})

    @asyncGen
    def web_duplicate(self):
        try:
            reqId = self.get_argument("ID")
            clearpp = self.get_argument("ClearPP")
            reqId = int(json.loads(reqId))
            clearpp = json.loads(clearpp)
        except Exception:
            self.finish({"success": "false", "result": [], "total": 0, "error": "Request ID is not a number"})

        result = yield self.threadTask(ProductionRequestClient().duplicateProductionRequest, reqId, clearpp)
        self.write(result)

    @asyncGen
    def web_delete(self):
        try:
            reqId = self.get_argument("ID")
            reqId = int(json.loads(reqId))
        except Exception:
            self.finish({"success": "false", "result": [], "total": 0, "error": "Request ID is not a number"})
        result = yield self.threadTask(ProductionRequestClient().deleteProductionRequest, reqId)
        self.write(result)

    @asyncGen
    def web_typeandmodels(self):
        types = [(x, x) for x in self.__request_types()]
        result = yield self.threadTask(
            ProductionRequestClient().getProductionRequestList, 0, "RequestID", "DESC", 0, 0, {"IsModel": 1}
        )
        models = []
        if not result["OK"]:
            models = []
        else:
            models = [
                (str(x["RequestID"]), str(x["RequestType"] + " - " + x["RequestName"])) for x in result["Value"]["Rows"]
            ]

        rows = [{"Name": x[0], "Description": x[1]} for x in types + models]
        self.finish({"OK": True, "result": rows, "total": len(rows)})

    @asyncGen
    def web_bkk_dq_list(self):
        result = yield self.threadTask(BookkeepingClient().getAvailableDataQuality)
        if not result["OK"]:
            self.finish({"success": "false", "result": [], "total": 0, "error": result["Message"]})
            return
        value = []
        if result["Value"]:
            value.append({"v": "ALL"})
        for x in result["Value"]:
            value.append({"v": x})
        self.finish({"OK": True, "total": len(value), "result": value})

    @asyncGen
    def web_bkk_event_types(self):
        addempty = "addempty" in self.request.arguments
        result = yield self.threadTask(BookkeepingClient().getAvailableEventTypes)
        if not result["OK"]:
            self.finish({"success": "false", "result": [], "total": 0, "error": result["Message"]})
            return
        rows = []
        for et in result["Value"]:
            rows.append({"id": et[0], "name": et[1], "text": "%s - %s" % (str(et[0]), str(et[1]))})
        rows.sort(key=lambda x: x["id"])
        if addempty:
            rows.insert(0, {"id": 99999999, "name": "", "text": "&nbsp;"})
        self.finish({"OK": True, "result": rows, "total": len(rows)})

    @asyncGen
    def web_bkk_input_tcks(self):
        pars = {
            "EventTypeId": self.get_argument("eventType", ""),
            "ConfigVersion": self.get_argument("configVersion", ""),
            "ProcessingPass": "/" + self.get_argument("inProPass", ""),
            "ConfigName": self.get_argument("configName", ""),
            "ConditionDescription": self.get_argument("simDesc", ""),
        }
        # log.info(str(pars))
        result = yield self.threadTask(BookkeepingClient().getTCKs, pars)
        if not result["OK"]:
            self.finish({"success": "false", "result": [], "total": 0, "error": result["Message"]})
            return
        gLogger.info(result["Value"])
        value = []
        if result["Value"]:
            value.append({"id": 0, "text": "ALL"})
        for x in result["Value"]:
            value.append({"id": str(x), "text": str(x)})
        value.sort(key=lambda x: x["id"])
        self.finish({"OK": True, "total": len(value), "result": value})

    @asyncGen
    def web_bkk_input_prod(self):
        pars = {
            "EventTypeId": self.get_argument("eventType", ""),
            "ConfigVersion": self.get_argument("configVersion", ""),
            "ProcessingPass": "/" + self.get_argument("inProPass", ""),
            "ConfigName": self.get_argument("configName", ""),
            "ConditionDescription": self.get_argument("simDesc", ""),
        }
        gLogger.info(str(pars))
        result = yield self.threadTask(BookkeepingClient().getProductions, pars)
        if not result["OK"]:
            self.finish({"success": "false", "result": [], "total": 0, "error": result["Message"]})
            return
        value = []
        if result["Value"]["Records"]:
            value.append({"id": 0, "text": "ALL"})
        for x in result["Value"]["Records"]:
            prod = x[0]
            if prod < 0:
                prod = -prod
            value.append({"id": prod, "text": prod})
        value.sort(key=lambda x: x["id"])
        self.finish({"OK": True, "total": len(value), "result": value})

    @asyncGen
    def web_bkk_simcond(self):
        result = yield self.threadTask(BookkeepingClient().getSimConditions)
        if not result["OK"]:
            self.finish({"success": "false", "result": [], "total": 0, "error": result["Message"]})
            return
        rows = [dict(zip(self.bkSimCondFields, sc)) for sc in result["Value"]]
        # !!! Sorting and selection must be moved to MySQL/Service side
        result = SelectAndSort(self.request, rows, "simCondID")
        self.finish(result)

    @asyncGen
    def web_getSimCondTree(self):
        _path, rtype, tree, dataQuality = self.__parseRequest()
        node = self.get_argument("node", "")

        bk = LHCB_BKKDBClient(web=True)

        yield self.threadTask(bk.setFileTypes, [])

        bk.setAdvancedQueries(rtype)
        bk.setParameter(tree)
        bk.setDataQualities(dataQuality)

        retVal = yield self.threadTask(bk.list, node)

        nodes = []
        if retVal:
            for i in retVal:
                node = {}
                node["text"] = i["name"]
                node["fullpath"] = i["fullpath"]
                node["id"] = i["fullpath"]
                node["selection"] = i["selection"] if "selection" in i else ""
                node["method"] = i["method"] if "method" in i else ""
                node["cls"] = "folder" if i["expandable"] else "file"
                if "level" in i and i["level"] == "Simulation Conditions/DataTaking":
                    node["cls"] = "file"
                    node["leaf"] = True
                else:
                    node["leaf"] = not i["expandable"]
                if "level" in i:
                    node["level"] = i["level"]
                    node["qtip"] = i["level"]
                nodes += [node]

        result = tornado.escape.json_encode(nodes)
        self.finish(result)

    @asyncGen
    def web_bkk_input_tree(self):
        keep = ["simCondID", "condType", "inProPass", "evType"]
        node = self.get_argument("node", "")

        bkcli = LHCB_BKKDBClient(web=True)

        value = []
        level = ""
        scd = {}
        known = {}

        nodes = yield self.threadTask(bkcli.list, node)
        for item in nodes:
            ipath = item["fullpath"]
            p = ipath.split("/")[1:]
            x = {"id": ipath, "text": p[-1]}
            x.update(dict(zip(["configName", "configVersion"], p[:2])))
            for i in keep:
                if self.get_argument(i, ""):
                    x[i] = self.get_arguments(i)
            if "level" in item:
                level = item["level"]
            if level == "Simulation Conditions/DataTaking":
                result = self.__condFromPath(ipath, bkcli, item)
                if not result["OK"]:
                    raise WErr.fromSERROR(result)
                x.update(result["Value"])
            elif level == "Processing Pass":
                if x.get("inProPass", ""):
                    x["inProPass"] = x["inProPass"][-1] + "/" + p[-1]
                else:
                    x["inProPass"] = p[-1]
            elif level == "Event types":
                x["evType"] = p[-1]
            elif level == "FileTypes":
                x["leaf"] = True
                x["inFileType"] = p[-1]
                if not scd:
                    result = self.__condFromPath(ipath, bkcli, {})
                    if not result["OK"]:
                        raise WErr.fromSERROR(result)
                    scd = result["Value"]
                x.update(scd)
            if x["text"] in known:
                gLogger.error("Duplicated entry: %s %s" % (str(known[x["text"]]), str(x)))
            else:
                known[x["text"]] = x
                value.append(x)
        result = tornado.escape.json_encode(value)
        self.finish(result)

    @asyncGen
    def web_save(self):
        rdict = {k: self.get_arguments(k) for k in self.request.arguments}
        self.__unescape(rdict)

        # Make it backward compatible
        for x in ["inTCKs"]:
            if x in rdict and str(rdict[x]) == "0":
                del rdict[x]
        reqId = ""
        if "ID" in rdict:
            try:
                reqId = int(rdict["ID"])
            except Exception:
                self.finish({"success": "false", "result": [], "total": 0, "error": "Request ID is not a number"})
                return
            del rdict["ID"]
            req = self.__fromlocal(rdict)
            result = yield self.threadTask(ProductionRequestClient().updateProductionRequest, reqId, req)
        else:
            sData = self.getSessionData()

            user = sData["user"]["username"]

            rdict["RequestAuthor"] = user
            req = self.__fromlocal(rdict)
            result = yield self.threadTask(ProductionRequestClient().createProductionRequest, req)

        if result["OK"]:
            self.finish({"success": "true", "requestId": reqId})
        else:
            raise WErr.fromSERROR(result)
        # self.finish()

    @asyncGen
    def web_templates(self):
        ret = yield self.threadTask(ProductionRequestClient().getProductionTemplateList)
        if not ret["OK"]:
            raise WErr.fromSERROR(ret)

        value = []
        for x in ret["Value"]:
            if "_run.py" not in x["WFName"]:
                gLogger.error("Skipping invalid production request template name:", x["WFName"])
                continue
            x["Type"] = "Script"
            value.append(x)
        result = {"OK": True, "total": len(value), "result": value}

        self.finish(result)

    def getRequestFields(self):
        result = dict.fromkeys(self.localFields, None)
        result.update(dict.fromkeys(self.simcondFields))
        result.update(dict.fromkeys(self.extraFields))
        for x in self.proStepFields:
            for i in range(1, 20):
                result["p%d%s" % (i, x)] = None
        sData = self.getSessionData()

        result["userName"] = sData["user"]["username"]
        result["userDN"] = sData["user"]["DN"]
        return result

    @asyncGen
    def web_template_parlist(self):
        tpl_name = self.get_argument("tpl", "")
        result = yield self.threadTask(ProductionRequestClient().getProductionTemplate, tpl_name)
        if not result["OK"]:
            raise WErr.fromSERROR(result)
        text = result["Value"]
        tpl = PrTpl(text)
        rqf = self.getRequestFields()
        tpf = tpl.getParams()
        tpd = tpl.getDefaults()
        plist = []
        for x in tpf:
            if x not in rqf:
                plist.append({"par": x, "label": tpf[x], "value": tpd[x].split("^")[0], "default": tpd[x]})
        self.finish({"OK": True, "total": len(plist), "result": plist})

    def getProductionRequest(self, ids):
        result = ProductionRequestClient().getProductionRequest(ids)
        if not result["OK"]:
            return result
        rr = result["Value"]
        lr = {}
        for x in rr:
            lr[x] = self.__2local(rr[x])
        return S_OK(lr)

    def getProductionRequestList(self, parent):
        try:
            offset = int(self.__getArgument("start", 0))
            limit = int(self.__getArgument("limit", 0))
            sortlist = self.__getJsonArgument("sort", [])
            if sortlist:
                sortBy = str(sortlist[-1]["property"])
                sortBy = self.serviceFields[self.localFields.index(sortBy)]
                sortOrder = str(sortlist[-1]["direction"])
            else:
                sortBy = self.serviceFields[self.localFields.index("ID")]
                sortOrder = "DESC"

            filterOpts = {}
            for x, y in [
                ("typeF", "RequestType"),
                ("stateF", "RequestState"),
                ("authorF", "RequestAuthor"),
                ("idF", "RequestID"),
                ("modF", "IsModel"),
                ("wgF", "RequestWG"),
            ]:

                val = self.get_argument(x, "")

                if val != "":
                    val = list(json.loads(val[-1]))
                    if val:
                        if not isinstance(val[0], bool):
                            filterOpts[y] = ",".join(val)
                        else:
                            filterOpts[y] = val[-1]

            if "IsModel" in filterOpts:
                if filterOpts["IsModel"]:
                    filterOpts["IsModel"] = 1
                else:
                    filterOpts["IsModel"] = 0
        except Exception as e:
            raise WErr(404, "Badly formatted list request: %s" + str(e))

        result = ProductionRequestClient().getProductionRequestList(
            parent, sortBy, sortOrder, offset, limit, filterOpts
        )

        if not result["OK"]:
            raise WErr.fromSERROR(result)

        rows = [self.__2local(x) for x in result["Value"]["Rows"]]
        aet = BookkeepingClient().getAvailableEventTypes()
        if aet["OK"]:
            etd = {}
            for et in aet["Value"]:
                etd[str(et[0])] = str(et[1])
            for x in rows:
                if "eventType" in x:
                    x["eventText"] = etd.get(str(x["eventType"]), "")

        return {"OK": True, "result": rows, "total": result["Value"]["Total"]}

    @asyncGen
    def web_create_workflow(self):
        """!!! Note: 1 level parent=master assumed !!!"""
        rdict = {k: self.get_argument(k) for k in self.request.arguments}
        for x in ["RequestID", "Template", "Subrequests"]:
            if x not in rdict:
                raise WErr(404, "Required parameter %s is not specified" % x)

        try:
            reqId = int(rdict["RequestID"])
            tpl_name = self.get_argument("Template")
            operation = self.get_argument("Operation", "Simple")
            sstr = self.get_argument("Subrequests")
            if sstr:
                slist = [int(x) for x in sstr.split(",")]
            else:
                slist = []
            sdict = dict.fromkeys(slist, None)
            del rdict["RequestID"]
            del rdict["Template"]
            del rdict["Subrequests"]
            if "Operation" in rdict:
                del rdict["Operation"]
        except Exception as e:
            raise WErr(404, "Wrong parameters (%s)" % str(e))
        requests = []
        if "RequestIDs" in rdict:
            requests = json.loads(rdict["RequestIDs"])
        if requests:
            res = []
            for reqId in requests:
                reqId = int(reqId)
                retVal = yield self.threadTask(self.__createScript, rdict, reqId, operation, tpl_name, sdict)
                if retVal["OK"]:
                    res += [val for val in retVal["Value"]]
                else:
                    raise WErr(
                        404, "The following error occurred during the production creation: %s" % retVal["Message"]
                    )
                if operation == "ScriptPreview":
                    break
            self.finish(S_OK(res))
        else:
            retVal = yield self.threadTask(self.__createScript, rdict, reqId, operation, tpl_name, sdict)
            if not retVal["OK"]:
                raise WErr(404, "The following error occurred during the production creation: %s" % retVal["Message"])
            self.finish(retVal)

    def __createScript(self, rdict, reqId, operation, tpl_name, sdict):
        """
        This method is used to create production or view the script using a certain template
        :param dict rdict: contains the production parameters
        :param int reqId: request id
        :param str operation: script previre or generate the production
        :param str tpl_name: template name use to create the workflow
        :param dict sdict: sub request parameters
        :return: S_OK(requestID:int, Body:list}
        """
        ret = ProductionRequestClient().getProductionTemplate(tpl_name)
        if not ret["OK"]:
            raise WErr.fromSERROR(ret)
        tpl = PrTpl(ret["Value"])

        ret = self.getProductionRequest([reqId])
        if not ret["OK"]:
            raise WErr.fromSERROR(ret)
        rqdict = ret["Value"][reqId]
        dictlist = []
        if rqdict["_is_leaf"]:
            d = self.getRequestFields()
            d.update(rqdict)
            d.update(rdict)
            dictlist.append(d)
        else:
            if not sdict:
                raise WErr(404, "Subrequests are not specified (but required)")
            ret = self.getProductionRequestList(reqId)
            if not ret["OK"]:
                raise WErr.fromSERROR(ret)
            for x in ret["result"]:
                if x["ID"] not in sdict:
                    continue
                d = self.getRequestFields()
                d.update(rqdict)
                for y in x:
                    if x[y]:
                        d[y] = x[y]
                d.update(rdict)
                dictlist.append(d)
        success = []
        fail = []
        for x in dictlist:
            for y in x:
                if x[y] is None:
                    x[y] = ""
                else:
                    x[y] = str(x[y])
            body = tpl.apply(x)
            if operation != "Generate":
                success.append({"ID": x["ID"], "Body": body})
            else:
                try:
                    res = ProductionRequestClient().execProductionScript(body, "")
                    if res["OK"]:
                        success.append({"ID": x["ID"], "Body": res["Value"]})
                    else:
                        fail.append(str(x["ID"]))
                except Exception:
                    fail.append(str(x["ID"]))
            continue  # not working with WF DB for now
            # name = 'PRQ_%s' % x['ID']
            # wf = fromXMLString( tpl.apply( x ) )
            # wf.setName( name )
            # wf.setType( 'production/requests' )
            # try:
            #   wfxml = wf.toXML()
            #   result = yield self.threadTask( ProductionRequestClient().publishWorkflow, str( wfxml ), True )
            # except Exception, msg:
            #   result = {'OK':False, 'Message': str( msg )}
            # if not result['OK']:
            #   fail.append( str( x['ID'] ) )
            # else:
            #   success.append( str( x['ID'] ) )

        if fail:
            raise WErr(404, "Couldn't get results from %s." % (",".join(fail)))
        else:
            # return S_OK("Success with %s" % ','.join(success))
            return S_OK(success)

    @asyncGen
    def web_split(self):
        try:
            reqId = int(self.get_argument("ID"))
            slist = list(json.loads(self.get_argument("Subrequests")))
        except Exception as e:
            raise WErr(400, "Wrong parameters (%s)" % str(e))

        result = yield self.threadTask(ProductionRequestClient().splitProductionRequest, reqId, slist)
        self.finish(result)

    @asyncGen
    def web_progress(self):
        try:
            reqId = int(self.get_argument("RequestID", "0"))
        except Exception:
            raise WErr(500, "Request ID is not a number")
        result = yield self.threadTask(ProductionRequestClient().getProductionProgressList, reqId)
        if not result["OK"]:
            raise WErr.fromSERROR(result)
        self.finish({"OK": True, "result": result["Value"]["Rows"], "total": result["Value"]["Total"]})

    @asyncGen
    def web_add_production(self):
        try:
            requestID = int(self.get_argument("RequestID"))
            productionID = int(self.get_argument("ProductionID"))
        except Exception:
            raise WErr(500, "Incorrect Request or production ID")

        result = yield self.threadTask(self.__bkProductionProgress, productionID)
        bkEvents = 0 if not result["OK"] else result["Value"]
        result = yield self.threadTask(
            ProductionRequestClient().addProductionToRequest,
            {"ProductionID": productionID, "RequestID": requestID, "Used": 1, "BkEvents": bkEvents},
        )
        self.finish(result)

    @asyncGen
    def web_remove_production(self):
        try:
            productionID = int(self.get_argument("ProductionID"))
        except ValueError:
            raise WErr(400, "Production ID is not a number")

        result = yield self.threadTask(ProductionRequestClient().removeProductionFromRequest, productionID)
        self.finish(result)

    @asyncGen
    def web_use_production(self):
        try:
            productionID = int(self.get_argument("ProductionID"))
            used = bool(int(self.get_argument("Used", "0")))
        except ValueError:
            raise WErr(400, "Incorrect Production ID use Used flag")

        result = yield self.threadTask(ProductionRequestClient().useProductionForRequest, productionID, used)
        self.finish(result)

    @asyncGen
    def web_productions(self):
        try:
            prId = int(self.get_argument("RequestID"))
        except ValueError:
            raise WErr(400, "Request ID is not a number")

        result = yield self.threadTask(ProductionRequestClient().getProductionList, prId)
        if not result["OK"]:
            raise WErr.fromSERROR(result)
        self.finish({"OK": True, "result": result["Value"], "total": len(result["Value"])})

    @asyncGen
    def web_getRequest(self):
        stepId = self.get_argument("StepId")
        requests = ProductionRequestManagerHandler.__dataCache.get("allData")
        if not requests:
            retVal = yield self.threadTask(
                ProductionRequestClient().getProductionRequestList, 0, "", "", 0, 0, {"RequestType": "Simulation"}
            )
            if not retVal["OK"]:
                raise WErr.fromSERROR(retVal)
            requests = retVal["Value"].get("Rows", [])
            ProductionRequestManagerHandler.__dataCache.add("allData", 3600, requests)

        records = []
        for request in requests:
            if request.get("ProDetail") and str(stepId) in request.get("ProDetail", ""):
                record = {}
                record["RequestId"] = request["RequestID"]
                record["RequestName"] = request["RequestName"]
                record["RequestWG"] = request["RequestWG"]
                record["RequestAuthor"] = request["RequestAuthor"]
                records.append(record)
        self.finish(
            {
                "success": "true",
                "result": records,
                "total": len(records),
                "date": Time.dateTime().strftime("%Y-%m-%d %H:%M [UTC]"),
            }
        )

    @asyncGen
    def web_getFastSimulationOpts(self):
        """

        This method used by the "Fast simulation type" combo box. The values of the combo box is filled using the CS

        """

        fastsimTypes = ["None", "Redecay"]
        csS = PathFinder.getServiceSection("ProductionManagement/ProductionRequest")
        if csS:
            fastsimTypes = gConfig.getValue("%s/FastSimulationType" % csS, fastsimTypes)

        rows = [{"text": i, "name": i} for i in fastsimTypes]
        self.finish({"OK": True, "result": rows, "total": len(rows)})



###############################################################################
# (c) Copyright 2020 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import defaultdict
import sys
import os
import re
import json
import zlib

import cachetools
import tornado
from tornado.escape import json_decode, json_encode

from DIRAC import gConfig, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Workflow.Workflow import fromXMLString as workflowFromXMLString
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import makeBKPath
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.Core.Utilities.ProductionData import getLogPath
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
from WebAppDIRAC.Lib.WebHandler import WebHandler, WErr, asyncGen


KNOWN_DATAPKGS = ["AnalysisProductions", "WG/CharmWGProd", "WG/CharmConfig"]
DONE_TRANSFORMATION_STATES = ["Archived", "Deleted", "Cleaned", "Completed"]
RE_PACKAGE_VERSION = re.compile(r"^v(?P<major>\d+)r(?P<minor>\d+)(?:p(?P<patch>\d+))?$")


class AnalysisProductionsHandler(WebHandler):
    AUTH_PROPS = "authenticated"

    _cache = cachetools.TTLCache(
        maxsize=10 * 1024 * 1024,  # 10 megabytes
        ttl=600,  # 10 minutes
        getsizeof=sys.getsizeof,
    )

    @classmethod
    def _getFromCache(cls, key):
        """Get a JSON response string from the lifetime based cache

        :param key: The key corresponding to this response
        :returns: The value from the cache or None
        """
        try:
            value = cls._cache[key]
        except KeyError:
            return None
        else:
            gLogger.debug("Found result in cache", str(key))
            return zlib.decompress(value).decode()

    @classmethod
    def _addToCache(cls, key, value):
        """Add a JSON response string to the cache

        The string will be stored in a compressed form to minimise memory usage.

        :param key: The key corresponding to this response
        :param value: The value corresponding to this response
        :returns: The value from the cache or None
        """
        cls._cache[key] = zlib.compress(value.encode())
        gLogger.debug("Added result to cache", "(current size = %d) %s" % (cls._cache.currsize, key))

    @asyncGen
    def web_getProduction(self):
        # Parse the request parameters

        requestIDs = self.get_argument("requestIDs", None)
        if not requestIDs:
            self.set_status(400)
            return self.finish(json_encode("Invalid value of requestIDs"))
        requestIDs = json_decode(requestIDs[0])
        if not requestIDs:
            return self.finish(json_encode([]))

        # Use the cache if possible
        result = self._getFromCache(frozenset(requestIDs))
        if result is not None:
            return self.finish(result)

        # Get the requested productions
        productions = getProductions({"RequestID": ",".join(map(str, requestIDs))})
        if set(requestIDs) != set(p["RequestID"] for p in productions):
            self.set_status(400)
            return self.finish(json_encode("Invalid request IDs passed"))

        # Collect all transformations in a single call (it's considerably faster)
        productionIDs = [p["RequestID"] for p in productions]
        result = TransformationClient().getTransformations({"TransformationFamily": productionIDs}, limit=1000)
        allTransformations = defaultdict(list)
        for transformation in _unwrap(result):
            productionID = int(transformation["TransformationFamily"])
            allTransformations[productionID] += [transformationToDict(transformation)]

        # Create dictionaries from each production request
        result = json_encode(
            [productionToDict(production, allTransformations[production["RequestID"]]) for production in productions]
        )

        self._addToCache(frozenset(requestIDs), result)
        return self.finish(result)

    @asyncGen
    def web_listProductions(self):
        # Use the cache if possible
        result = self._getFromCache("listProductions")
        if result is not None:
            return self.finish(result)

        # Find the production requests
        rows = []
        rows += getProductions({"RequestType": "AnalysisProduction"})
        # This search filter isn't ideal but it's the only way of efficiently
        # querying for legacy productions
        rows += getProductions({"RequestType": "WGProduction", "RequestAuthor": "cburr"})

        # Transform them into a tree
        # Structure is WG > Name > Version > List[Requests]
        productions = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
        for row in rows:
            package, version = findPackageNameAndVersion(row["ProDetail"])

            wg = row["RequestWG"]
            name = row["RequestName"]

            # This is to account for legacy productions from WG/CharmWGProd
            # TODO: Modify the database to remove the need for this hack?
            if package != "AnalysisProductions":
                version += " (" + package + ")"
                wg = "Unknown"
                optionsDirectories = [
                    "CHARMWGPRODROOT/productions/",
                    "CHARMWGPRODROOT/../productions/",
                    "CHARMCONFIGROOT/options/WGProductions/",
                    "CHARMCONFIGOPTS/WGProductions/",
                ]
                for s in optionsDirectories:
                    if s in row["ProDetail"]:
                        name = row["ProDetail"].split(s)[1].split("/")[0]
                        break
                else:
                    raise NotImplementedError(row.get("RequestID"))

            productions[wg][name][version].append({"id": row["RequestID"], "archived": False, "published": False})

        productions = _covertToTreeModel(productions)
        result = json_encode(productions)
        self._addToCache("listProductions", result)
        return self.finish(result)


def _covertToTreeModel(data, name="."):
    """Recursively prepare data for use with an ExtJS TreeModel"""
    node = {"name": name}

    if isinstance(data, list):
        child_data = data
        node["requestIDs"] = json_encode([c["id"] for c in child_data])
        node["leaf"] = True
    else:
        child_data = [_covertToTreeModel(v, name=k) for k, v in data.items()]
        node["children"] = sorted(child_data, key=_nodeToSortKey)

    node["archived"] = all(c["archived"] for c in child_data)
    node["published"] = any(c["published"] for c in child_data)

    return node


def _unwrap(result):
    """Unwrap an S_OK/S_ERROR response, raising if appropriate

    :param result: The response from a DIRAC RPC call
    :returns: The value returned by the RPC call
    :raises: :exc:`WebAppDIRAC.Lib.WebHandler.WErr`
    """
    if not result["OK"]:
        raise WErr.fromSERROR(result)
    return result["Value"]


def _nodeToSortKey(node):
    """Key function for use with "sorted" which understands version numbers

    :param node: The extJS node data to compure the key for
    :returns: Either a string or a tuple of integers
    """
    match = RE_PACKAGE_VERSION.match(node["name"].split(" ")[0])
    if match:
        # Return negative numbers to ensure the most recent is at the top
        return [-int(x) if x else 0 for x in match.groups()]
    else:
        return node["name"]


def getProductionStatus(production):
    """Find the analyst readable production status

    Converts a dictionary describing a production (from productionToDict) into
    a string summarising the status of the production. The allowed values are:

      * waiting: Waiting for the production to start
      * active: The transformation system is processing the data
      * replicating: Waiting for the data to be copied to CERN
      * ready: Ready for use
      * archived: This has been replaced by a newer production
      * published: This has been used for a publication

    :param production: A dictionary describing a production from productionToDict
    :returns: A string describing the production's state
    """
    if production["archived"]:
        return "archived"

    if production["published"]:
        return "published"

    if not production["transformations"]:
        return "waiting"

    transformationActive = False
    for transformation in production["transformations"]:
        if transformation["transformationStatus"] in DONE_TRANSFORMATION_STATES:
            pass
        else:
            fileCounts = transformation["fileCounts"]
            if fileCounts.get("Processed", 0) != fileCounts["Total"]:
                transformationActive = True
    if transformationActive:
        return "active"

    if production["numOutputFiles"] != production["numOutputFilesAtCERN"]:
        return "replicating"

    return "ready"


def getProductions(query):
    """Query the production management system for a list of productions

    :param query: A dictionary of query parameters to pass to getProductionRequestList
    :returns: A list which each element corresponding to a row in the database
    """
    result = ProductionRequestClient().getProductionRequestList(
        0,  # Parent
        "RequestID",  # Sortby
        "DESC",  # Sort order
        0,  # Offset
        0,  # Max results
        query,
    )
    if result["OK"]:
        return _unwrap(result)["Rows"]
    else:
        raise WErr.fromSERROR(result)


def findPackageNameAndVersion(prodDetail):
    """Find the data package and version which was used to submit this production

    :param prodDetail: The prodDetail field of the production request
    :returns: A tuple containing two strings (packageName, version)
    """
    # Use regex for now instead of handling unpickling the string
    # Look for patterns like "AnalysisProductions.v1r2p3" and "CharmWGProd.v1r5"
    # TODO: Use the JSONOrPickle function when this is in v10r1
    pattern = re.compile(r"(" + "|".join([s.split("/")[-1] for s in KNOWN_DATAPKGS]) + ")" r"\.(v\d+r\d+(?:p\d+)?)")
    submitPackage = set(pattern.findall(prodDetail))
    if len(submitPackage) != 1:
        raise NotImplementedError(prodDetail, submitPackage)
    return submitPackage.pop()


def productionToDict(production, transformations):
    """Convert the response of getProductionRequestList to a dictionary

    :param production: A row returned from getProductionRequestList
    :param transformations: A list of transformations associated with this production
    :returns: A dictionary summarising this production
    """
    productionID = production["RequestID"]

    if production["RequestType"] == "AnalysisProduction":
        _, version = findPackageNameAndVersion(production["ProDetail"])
        proDetail = json.loads(production["ProDetail"])
        productionName = proDetail["p1Name"].split("-" + version + "-")[-1].split("/")[0].strip()
    else:
        productionName = "-".join(production["RequestName"].split("-")[1:])

    lfns = []
    pfns = []
    pfnsAtCERN = []
    outputSize = 0
    outputSizeAtCERN = 0
    if transformations:
        transformationID = transformations[-1]["transformationID"]

        lfnMetadata = _unwrap(BookkeepingClient().getProductionFiles(transformationID, "ALL", "ALL"))
        # Find the LFNs which were intended as output and their size
        # Mostly needed to ignore LOG files
        expectedFileTypes = transformations[-1]["steps"][-1]["outputTypes"]
        for lfn, metadata in lfnMetadata.items():
            if metadata["FileType"] not in expectedFileTypes:
                continue
            if not metadata["GotReplica"].upper().startswith("Y"):
                continue
            lfns.append(lfn)
            outputSize += metadata["FileSize"]

        if lfns:
            replicas = _unwrap(DataManager().getReplicas(lfns, getUrl=False))["Successful"]
            # Find the replicas at CERN
            for se in gConfig.getValue("/Resources/Sites/LCG/LCG.CERN.cern/SE", []):
                seLFNs = [lfn for lfn in replicas if se in replicas[lfn]]
                if not seLFNs:
                    continue
                # Lookup the corresponding PFNs
                result = _unwrap(StorageElement(se).getURL(seLFNs, protocol="root"))
                pfnsAtCERN += list(result["Successful"].values())
                # Compute the size in bytes of the data at CERN
                outputSizeAtCERN += sum(lfnMetadata[lfn]["FileSize"] for lfn in seLFNs)

    simCondDetail = json.loads(production["SimCondDetail"])
    productionInputQuery = os.path.join(
        "/",
        simCondDetail["configName"],
        simCondDetail["configVersion"],
        production["SimCondition"],
        simCondDetail["inProPass"],
        production["EventType"],
        simCondDetail["inFileType"],
    )

    outputPattern = ""
    if pfnsAtCERN:
        # Get the file extension
        extensions = set(os.path.splitext(p)[1] for p in pfnsAtCERN)
        if len(extensions) > 1:
            raise NotImplementedError(extensions)
        suffix = "*" + extensions.pop()

        # Find a common directory
        pfnDirectories = set(os.path.dirname(p) for p in pfnsAtCERN)
        if len(pfnDirectories) > 1:
            pfnDirectories = set(os.path.dirname(p) for p in pfnsAtCERN)
            if len(pfnDirectories) > 1:
                # This only goes two levels deep by convention
                raise NotImplementedError(pfnDirectories)
            suffix = "*/" + suffix
        outputPattern = pfnDirectories.pop() + "/" + suffix

    result = {
        "requestID": productionID,
        "status": production["RequestState"],
        "name": productionName,
        "outputPattern": outputPattern,
        "LFNs": lfns,
        "PFNs": pfnsAtCERN,
        "outputSize": outputSize,
        "numOutputFiles": len(lfns),
        "numOutputFilesAtCERN": len(pfnsAtCERN),
        "productionInputQuery": productionInputQuery,
        "transformations": transformations,
        "archived": False,  # TODO
        "published": False,  # TODO
    }
    result["state"] = getProductionStatus(result)
    return result


def transformationToDict(transformation):
    """Convert the response of getTransformations to a dictionary

    :param production: A row returned from getTransformations
    :returns: A dictionary summarising this transformation
    """
    transformationID = transformation["TransformationID"]
    transformationStatus = transformation["Status"]

    workflow = workflowFromXMLString(transformation["Body"])
    parameters = {p.getName(): p.getValue() for p in workflow.parameters}

    # Skip this if the transfromation is done as it's relatively slow
    fileCounts = {}
    if transformationStatus not in DONE_TRANSFORMATION_STATES:
        fileCounts = _unwrap(TransformationClient().getTransformationFilesCount(transformationID, "Status"))

    logsURL = getTransformationLogsURL(parameters["configName"], parameters["configVersion"], transformationID)

    steps = [
        stepToDict(parameters["BKProcessingPass"]["Step" + str(i)]) for i in range(len(parameters["BKProcessingPass"]))
    ]

    transformInputQuery = _unwrap(TransformationClient().getBookkeepingQuery(transformationID))

    return {
        "transformationID": transformationID,
        "transformationStatus": transformation["Status"],
        "transformationType": transformation["Type"],
        "fileCounts": fileCounts,
        "logsURL": logsURL,
        "steps": steps,
        "transformInputQuery": transformInputQuery,
    }


def stepToDict(stepData):
    """Convert a BKProcessingPass step to a dictionary

    :param production: A step from the BKProcessingPass parameter of a workflow
    :returns: A dictionary summarising this step
    """
    extras = stepData["ExtraPackages"].split(";")
    for extra in extras:
        if any(extra.startswith(package) for package in KNOWN_DATAPKGS):
            package, version = extra.split(".")
            if not RE_PACKAGE_VERSION.match(version):
                raise NotImplementedError(package, version)
            gitlabURL = "https://gitlab.cern.ch/lhcb-datapkg/" + package + "/-/tree/" + version
            break
    else:
        gitlabURL = None

    outputTypes = [d["FileType"] for d in stepData["OutputFileTypes"]]

    return {
        "stepID": stepData["BKStepID"],
        "application": stepData["ApplicationName"] + "/" + stepData["ApplicationVersion"],
        "extras": extras,
        "GitLabURL": gitlabURL,
        "options": stepData["OptionFiles"].split(";"),
        "outputTypes": outputTypes,
    }


def getTransformationLogsURL(configName, configVersion, transformationID):
    """Get the HTTPS URL on logSE for a given transformation

    :param configName: The transformations config name
    :param configVersion: The transformations config version
    :param transformationID: The ID of the transformation
    :returns: A string corresponding to the base URL of the logs
    """
    from LHCbDIRAC.Core.Utilities.ProductionData import _makeProductionPath, _getLFNRoot

    lfnRoot = _getLFNRoot("", configName, configVersion)
    lfn = _makeProductionPath("", lfnRoot, "LOG", str(transformationID).zfill(8))
    logSE = Operations().getValue("LogStorage/LogSE", "LogSE")
    result = StorageElement(logSE).getURL(lfn, protocol="https")
    return _unwrap(result)["Successful"][lfn]



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import re
import json

from DIRAC import S_ERROR


class PrTpl(object):
    """Production Template engine
    AZ: I know it is not the best :)
    It undestand: {{<sep><ParName>[#<Label>[#<DefaultValue>]]}}
    Where:
      <sep> - not alpha character to be inserted
              before parameter value in case it
              is not empty
      <ParName> - parameter name (must begin
              with alpha character) to substitute
      <Label> - if '#' is found after '{{', everything after
              it till '#' or '}}' is the Label for the parameter
              (default to ParName) In case you use one parameter
              several times, you can specify the label only once.
      <DefaultValue> - the value in the disalog by default
              (empty string if not specified)
    """

    __par_re = r"\{\{(\W)?(.+?)(|(#.*?))\}\}"
    __par_i_re = r"#([^\#]*)#?(.*)"
    __par_sub = r"\{\{(\W|)?%s(|(#.*?))\}\}"

    def __init__(self, tpl_xml):
        self.tpl = tpl_xml
        self.pdict = {}
        self.ddict = {}
        for x in re.findall(self.__par_re, self.tpl):
            rest = re.findall(self.__par_i_re, x[3])
            if rest:
                rest = rest[0]
            else:
                rest = ("", "")
            if not x[1] in self.ddict or rest[1]:
                self.ddict[x[1]] = rest[1]
            if x[1] in self.pdict and not rest[0]:
                continue
            label = rest[0]
            if not label:
                label = x[1]
            self.pdict[x[1]] = label

    def getParams(self):
        """Return the dictionary with parameters (value is label)"""
        return self.pdict

    def getDefaults(self):
        """Return the dictionary with parameters defaults"""
        return self.ddict

    def apply(self, pdict):
        """Return template with substituted values from pdict"""
        result = self.tpl
        for p in self.pdict:
            value = str(pdict.get(p, ""))
            if value:
                value = "\\g<1>" + value
            result = re.sub(self.__par_sub % p, value, result)
        return result


def SelectAndSort(request, rows, default_sort):
    """Can be used to return extJS adopted list of rows"""
    total = len(rows)
    if total == 0:
        return {"OK": True, "result": [], "total": 0}
    try:
        start = int(request.arguments.get("start", 0)[-1])

        if start != 0:
            start = int(json.loads(request.arguments.get("start")[-1]))

        limit = int(request.arguments.get("limit", total)[-1])
        if limit != total:
            limit = int(json.loads(request.arguments.get("limit", total)[-1]))

        sortList = request.arguments.get("sort", default_sort)[-1]
        dir = "ASC"
        if sortList != default_sort:
            sortList = json.loads(request.arguments.get("sort", default_sort)[-1])
            dir = sortList[0]["direction"]
            sort = sortList[0]["property"]

        if sort not in rows[0]:
            raise Exception("Sort field " + sort + " is not found")
    except Exception as e:
        return S_ERROR("Badly formatted list request: " + str(e))

    rows.sort(key=lambda x: x[sort], reverse=(dir != "ASC"))
    return {"OK": True, "result": rows[start : start + limit], "total": total}



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import json
import tornado
import sys
import datetime
import tempfile
import tarfile
import os
import shutil

from DIRAC import gLogger
from DIRAC.Core.Utilities import Time
from DIRAC.FrameworkSystem.Client.UserProfileClient import UserProfileClient
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from LHCbDIRAC.Interfaces.API.DiracLHCb import DiracLHCb
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager

from LHCbDIRAC.BookkeepingSystem.Client.LHCB_BKKDBClient import LHCB_BKKDBClient
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

from WebAppDIRAC.Lib.WebHandler import WebHandler, asyncGen, WErr


class BookkeepingBrowserHandler(WebHandler):

    AUTH_PROPS = "authenticated"

    numberOfJobs = None
    pageNumber = None

    def index(self):
        pass

    @asyncGen
    def web_getNodes(self):

        _, querytype, tree, dataQuality = self.__parseRequest()
        node = self.get_argument("node", "")

        bk = LHCB_BKKDBClient(web=True)

        yield self.threadTask(bk.setFileTypes, [])

        bk.setAdvancedQueries(querytype)
        bk.setParameter(tree)
        bk.setDataQualities(dataQuality)

        retVal = yield self.threadTask(bk.list, node)

        nodes = []
        if len(retVal) > 0:
            for i in retVal:
                node = {}
                node["text"] = i["name"]
                node["fullpath"] = i["fullpath"]
                node["id"] = i["fullpath"]
                node["selection"] = i["selection"] if "selection" in i else ""
                node["method"] = i["method"] if "method" in i else ""
                node["cls"] = "folder" if i["expandable"] else "file"
                if "level" in i and i["level"] == "Event types":
                    node["text"] = "%s (%s)" % (i["name"], i["Description"])
                if "level" in i and i["level"] == "FileTypes":
                    node["leaf"] = True
                else:
                    node["leaf"] = False if i["expandable"] else True
                if "level" in i:
                    node["level"] = i["level"]
                    node["qtip"] = i["name"]
                nodes += [node]

        result = tornado.escape.json_encode(nodes)
        self.finish(result)

    @asyncGen
    def web_getdataquality(self):
        bk = LHCB_BKKDBClient(web=True)
        result = yield self.threadTask(bk.getAvailableDataQuality)
        if result["OK"]:
            ret = []
            for i in result["Value"]:
                checked = True if i == "OK" else False
                ret += [{"name": i, "value": checked}]
            self.finish({"success": "true", "result": ret})
        else:
            self.finish({"result": [], "error": result["Message"]})

    @asyncGen
    def web_getFiles(self):

        path, querytype, tree, dataQuality = self.__parseRequest()

        bk = LHCB_BKKDBClient(web=True)

        bk.setAdvancedQueries(querytype)
        bk.setParameter(tree)
        bk.setDataQualities(dataQuality)

        retVal = yield self.threadTask(
            bk.getLimitedFiles, {"fullpath": path}, {"total": "0"}, self.pageNumber, self.numberOfJobs + self.pageNumber
        )

        if not retVal["OK"]:
            raise WErr.fromSERROR(retVal)
        nbrecords = retVal["Value"]["TotalRecords"]
        if nbrecords > 0:
            params = retVal["Value"]["ParameterNames"]
            records = []
            for i in retVal["Value"]["Records"]:
                k = [j if j and j != "None" else "-" for j in i]
                records += [dict(zip(params, k))]
            extras = {}
            if "Extras" in retVal["Value"]:
                extras = retVal["Value"]["Extras"]
                extras["GlobalStatistics"]["Number of Files"] = nbrecords
                size = self.__bytestr(extras["GlobalStatistics"]["Files Size"])
                extras["GlobalStatistics"]["Files Size"] = size

            timestamp = Time.dateTime().strftime("%Y-%m-%d %H:%M [UTC]")
            data = {
                "success": "true",
                "result": records,
                "date": timestamp,
                "total": nbrecords,
                "ExtraParameters": extras,
            }
        else:
            data = {"success": "false", "result": [], "error": "Nothing to display!"}

        self.finish(data)

    ################################################################################
    @staticmethod
    def __bytestr(size, precision=1):
        """Return a string representing the greek/metric suffix of a size"""
        abbrevs = [
            (1 << 50, " PB"),
            (1 << 40, " TB"),
            (1 << 30, " GB"),
            (1 << 20, " MB"),
            (1 << 10, " kB"),
            (1, " bytes"),
        ]
        if size is None:
            return "0 bytes"
        if size == 1:
            return "1 byte"
        factor = None
        suffix = None
        for factor, suffix in abbrevs:
            if size >= factor:
                break
        float_string_split = repr(size / float(factor)).split(".")
        integer_part = float_string_split[0]
        decimal_part = float_string_split[1]
        if int(decimal_part[0:precision]):
            float_string = ".".join([integer_part, decimal_part[0:precision]])
        else:
            float_string = integer_part
        return float_string + suffix

    def __parseRequest(self):
        path = self.get_argument("fullpath", "")
        querytype = self.get_argument("type", "") == "adv"
        tree = self.get_argument("tree", "Configuration")

        dataQuality = None
        if "dataQuality" in self.request.arguments:
            dataQuality = dict(json.loads(self.get_argument("dataQuality")))
            if not dataQuality:
                dataQuality = {"OK": True}

        self.numberOfJobs = int(self.get_argument("limit", "25"))
        self.pageNumber = int(self.get_argument("start", "0"))

        return path, querytype, tree, dataQuality

    @asyncGen
    def web_getStatistics(self):

        path, querytype, tree, dataQuality = self.__parseRequest()

        bk = LHCB_BKKDBClient(web=True)

        bk.setAdvancedQueries(querytype)
        bk.setParameter(tree)
        bk.setDataQualities(dataQuality)

        retVal = yield self.threadTask(
            bk.getLimitedInformations, self.pageNumber, self.numberOfJobs + self.pageNumber, path
        )
        if retVal["OK"]:
            value = {}
            value["nbfiles"] = retVal["Value"]["Number of files"]
            value["nbevents"] = self.__niceNumbers(retVal["Value"]["Number of Events"])
            value["fsize"] = self.__bytestr(retVal["Value"]["Files Size"])
            data = {"success": "true", "result": value}
        else:
            data = {"success": "false", "result": [], "error": retVal["Message"]}
        # data = {"success":"true","result":{'nbfiles':0,'nbevents':0,'fsize':0}}
        self.finish(data)

    @staticmethod
    def __niceNumbers(number):
        strList = list(str(number))
        newList = [strList[max(0, i - 3) : i] for i in range(len(strList), 0, -3)]
        newList.reverse()
        finalList = []
        for i in newList:
            finalList.append(str("".join(i)))
        finalList = " ".join(map(str, finalList))
        return finalList

    @asyncGen
    def web_saveDataSet(self):

        path, querytype, tree, dataQuality = self.__parseRequest()

        bk = LHCB_BKKDBClient(web=True)

        bk.setAdvancedQueries(querytype)
        bk.setParameter(tree)
        bk.setDataQualities(dataQuality)

        fileformat = self.get_argument("format", None)
        fileName = self.get_argument("fileName", "bookkeeping_files")
        if not fileName.lower().endswith("." + fileformat.lower()):
            fileName += "." + fileformat.lower()

        if fileformat in ("py", "txt"):
            try:
                data = yield self.threadTask(
                    bk.writePythonOrJobOptions, self.pageNumber, self.numberOfJobs + self.pageNumber, path, fileformat
                )
            except Exception:
                data = {"success": "false", "error": str(sys.exc_info()[1])}
        elif "bkQuery" in self.request.arguments:
            bkQuery = dict(json.loads(self.get_argument("bkQuery")))
            retVal = yield self.threadTask(bk.getFilesWithMetadata, bkQuery)
            if not retVal["OK"]:
                data = {"success": "false", "error": retVal["Message"]}
            else:
                fileContent = [",".join(retVal["Value"]["ParameterNames"])]
                for record in retVal["Value"]["Records"][self.pageNumber : self.numberOfJobs + self.pageNumber]:
                    fileContent += [",".join(str(metadata) for metadata in record)]
                data = "\n".join(fileContent)
        else:
            self.finish({"success": "false", "error": "Please provide fileName and format!"})

        self.set_header("Content-type", "text/plain")
        self.set_header("Content-Disposition", 'attachment; filename="%s' % fileName)
        self.set_header("Content-Length", len(data))
        self.set_header("Content-Transfer-Encoding", "Binary")
        self.set_header("Cache-Control", "no-cache, no-store, must-revalidate, max-age=0")
        self.set_header("Pragma", "no-cache")
        self.set_header(
            "Expires", (datetime.datetime.utcnow() - datetime.timedelta(minutes=-10)).strftime("%d %b %Y %H:%M:%S GMT")
        )
        self.finish(data)

    @asyncGen
    def web_getBookmarks(self):
        upc = UserProfileClient("Bookkeeping")
        result = yield self.threadTask(upc.retrieveVar, "Bookmarks")
        if result["OK"]:
            data = []
            for i in result["Value"]:
                data += [{"name": i, "value": result["Value"][i]}]
            result = {"success": "true", "result": data}
        else:
            if result["Message"].find("No data for") != -1:
                result = {"success": "true", "result": {}}
            else:
                result = {"success": "false", "error": result["Message"]}
        self.finish(result)

    @asyncGen
    def web_addBookmark(self):
        title = self.get_argument("title", "")
        path = self.get_argument("path", "")

        upc = UserProfileClient("Bookkeeping")
        result = yield self.threadTask(upc.retrieveVar, "Bookmarks")
        data = result["Value"] if result["OK"] else {}
        if title in data:
            result = {"success": "false", "error": 'The bookmark with the title "' + title + '" is already exists'}
        else:
            data[title] = path
        result = yield self.threadTask(upc.storeVar, "Bookmarks", data, {"ReadAccess": "ALL"})
        if result["OK"]:
            result = {"success": "true", "result": "It successfully added to the bookmark!"}
        else:
            result = {"success": "false", "error": result["Message"]}
        self.finish(result)

    @asyncGen
    def web_deleteBookmark(self):
        title = self.get_argument("title", "")

        upc = UserProfileClient("Bookkeeping")
        result = yield self.threadTask(upc.retrieveVar, "Bookmarks")

        data = result["Value"] if result["OK"] else {}
        if title in data:
            del data[title]
        else:
            result = {"success": "false", "error": "Can't delete not existing bookmark: \"" + title + '"'}

        result = yield self.threadTask(upc.storeVar, "Bookmarks", data, {"ReadAccess": "ALL"})
        if result["OK"]:
            result = {"success": "true", "result": "It successfully deleted to the bookmark!"}
        else:
            result = {"success": "false", "error": result["Message"]}
        self.finish(result)

    @asyncGen
    def web_jobinfo(self):
        """
        For retrieving the job information for a given lfn
        """
        lfn = self.get_argument("lfn", None)

        bk = LHCB_BKKDBClient(web=True)
        result = yield self.threadTask(bk.getJobInfo, lfn)

        if result is None:
            result = {"success": "false", "error": "Can not retrive job information"}
        else:
            jobinfos = [[key, value] for key, value in result.items()]
            result = {"success": "true", "result": jobinfos}
        self.finish(result)

    @asyncGen
    def web_ancestors(self):
        """
        For retrieving the ancestors for a given lfn
        """
        lfn = self.get_argument("lfn", None)

        bk = LHCB_BKKDBClient(web=True)
        result = yield self.threadTask(bk.getFileHistory, lfn)
        if result["OK"]:
            files = []
            nbrecords = result["Value"]["TotalRecords"]
            if nbrecords > 0:
                params = result["Value"]["ParameterNames"]
                for i in result["Value"]["Records"]:
                    k = [str(j) if j and j != "None" else "-" for j in i]
                    files += [dict(zip(params, k))]
                result = {"success": "true", "result": files}
            else:
                result = {"success": "false", "error": "No ancestors found!"}
        else:
            WErr.fromSERROR(result)
            result = {"success": "false", "error": result["Message"]}
        self.finish(result)

    @asyncGen
    def web_logfile(self):
        """
        Retrieve the log files for a given lfn
        """
        lfn = self.get_argument("lfn", None)

        bk = LHCB_BKKDBClient(web=True)
        result = yield self.threadTask(bk.getLogfile, lfn)
        if result["OK"]:
            result = {"success": "true", "result": result["Value"]}
        else:
            result = {"success": "false", "error": "No log file found!"}
        self.finish(result)

    @asyncGen
    def web_processingpass(self):
        """
        Retrieve the processing pass for a given dataset.
        """
        stepname = self.get_argument("stepName", None)

        bk = LHCB_BKKDBClient(web=True)

        result = yield self.threadTask(bk.getProcessingPassSteps, {"StepName": stepname})
        if result["OK"]:
            # convert the data to the correct format
            records = result["Value"]["Records"]
            steps = [dict([step for step in records[record]]) for record in records]
            result = {"success": "true", "result": steps}
        else:
            result = {"success": "false", "error": result["Message"]}
        self.finish(result)

    @asyncGen
    def web_stepmetadata(self):
        """
        Retrieve the processing pass for a given dataset.
        """
        bkQuery = self.get_argument("bkQuery", None)
        if bkQuery is not None:
            bkQuery = dict(json.loads(bkQuery))

        bk = LHCB_BKKDBClient(web=True)

        result = yield self.threadTask(bk.getStepsMetadata, bkQuery)

        if result["OK"]:
            # convert the data to the correct format
            records = result["Value"]["Records"]
            steps = [dict([step for step in records[record]]) for record in records]
            result = {"success": "true", "result": steps}
        else:
            result = {"success": "false", "error": result["Message"]}
        self.finish(result)

    @asyncGen
    def web_conditions(self):
        """
        Retrieve the simulation or data taking conditions
        """
        bkQuery = self.get_argument("bkQuery", None)
        if bkQuery is not None:
            bkQuery = dict(json.loads(bkQuery))

        result = yield self.threadTask(BookkeepingClient().getConditions, bkQuery)

        if result["OK"]:
            # convert the data to the correct format
            records = result["Value"]
            conditions = []
            if records[0]["TotalRecords"] > 0:
                paramNames = records[0]["ParameterNames"]
                for rec in records[0]["Records"]:
                    conditions = dict(zip(paramNames, rec))
                    if bkQuery["ConditionDescription"] in conditions.values():
                        condType = "sim"
                        break
            else:
                paramNames = records[1]["ParameterNames"]
                for rec in records[1]["Records"]:
                    conditions = dict(zip(paramNames, rec))
                    if bkQuery["ConditionDescription"] in conditions.values():
                        condType = "daq"
                        break
            result = {"success": "true", "result": conditions, "CondType": condType}
        else:
            result = {"success": "false", "error": result["Message"]}
        self.finish(result)

    @asyncGen
    def web_runs(self):
        """
        For retrieving list of runs
        """

        bk = LHCB_BKKDBClient(web=True)
        bk.setParameter("Runlookup")

        retVal = yield self.threadTask(bk.list)

        data = []
        if len(retVal) > 0:
            for i in retVal:
                data.append({"data": i["name"]})
            result = {"success": "true", "result": sorted(data)}
        else:
            result = {"success": "false", "error": "no data found"}

        self.finish(result)

    @asyncGen
    def web_productions(self):
        """
        For retrieving list of runs
        """

        bk = LHCB_BKKDBClient(web=True)
        bk.setParameter("Productions")

        retVal = yield self.threadTask(bk.list)

        data = []
        if len(retVal) > 0:
            for i in retVal:
                data.append({"data": i["name"]})
            result = {"success": "true", "result": sorted(data)}
        else:
            result = {"success": "false", "error": "no data found"}

        self.finish(result)

    @asyncGen
    def web_t1sites(self):
        """
        Retrive the list of Tier1 sites
        """

        try:
            shortSiteNames = DMSHelpers().getShortSiteNames(withStorage=False, tier=(0, 1))
        except AttributeError:
            shortSiteNames = {
                "CERN": "LCG.CERN.cern",
                "RAL": "LCG.RAL.uk",
                "IN2P3": "LCG.IN2P3.fr",
                "GRIDKA": "LCG.GRIDKA.de",
                "NIKHEF": "LCG.NIKHEF.nl",
                "CNAF": "LCG.CNAF.it",
                "RRCKI": "LCG.RRCKI.ru",
                "PIC": "LCG.PIC.es",
            }
            # no we have to convert to the proper format
        data = []
        for shortSiteName, siteName in shortSiteNames.items():
            data.append({"Name": shortSiteName, "Value": siteName})

        self.finish({"success": "true", "result": sorted(data)})

    @asyncGen
    def web_createCatalog(self):
        """It is used for create pool xml catalog"""

        _, querytype, tree, dataQuality = self.__parseRequest()

        bk = LHCB_BKKDBClient(web=True)
        bk.setAdvancedQueries(querytype)
        bk.setParameter(tree)
        bk.setDataQualities(dataQuality)

        siteName = self.get_argument("SiteName")
        formatType = self.get_argument("formatType")
        fileName = self.get_argument("fileName")
        bkQuery = dict(json.loads(self.get_argument("bkQuery")))

        userName = self.getUserName()
        result = gProxyManager.downloadVOMSProxyToFile(
            self.getUserDN(), self.getUserGroup(), limited=True, requiredTimeLeft=86400, cacheTime=86400
        )

        if not result["OK"]:
            gLogger.error("Failed to set shifter proxy", result["Message"])
            self.finish(
                {
                    "success": "false",
                    "error": "Can not retrieve proxy used to generate the xml catalog! "
                    "Please upload your proxy to LHCbDIRAC!",
                }
            )
            return

        proxyFile = result["Value"]
        os.environ["X509_USER_PROXY"] = proxyFile

        retVal = yield self.threadTask(bk.getFilesWithMetadata, bkQuery)
        if not retVal["OK"]:
            self.finish({"success": "false", "error": retVal["Message"]})
            return

        lfns = {}
        for record in retVal["Value"]["Records"][self.pageNumber : self.numberOfJobs + self.pageNumber]:
            fileMetaDict = dict(zip(retVal["Value"]["ParameterNames"], record))
            lfns[fileMetaDict["FileName"]] = fileMetaDict
        if not lfns:
            self.finish({"success": "false", "error": "No files found"})
            return

        tmpdir = tempfile.mkdtemp(prefix=userName)
        catalog = "%s/%s.xml" % (tmpdir, fileName)
        retVal = yield self.threadTask(DiracLHCb().getInputDataCatalog, list(lfns), siteName, catalog, True)
        if not retVal["OK"]:
            raise WErr.fromSERROR(retVal)

        slist = retVal["Value"].get("Successful", {})
        exists = {}
        for lfn in slist.keys():
            exists[lfn] = lfns[lfn]

        if not fileName.endswith(".py"):
            fileName = "%s/%s.py" % (tmpdir, fileName)

        kwargs = dict(catalog=catalog, dataset=bkQuery)
        if formatType.lower() == "pfn":
            kwargs = dict(savedType=None, savePfn=slist)
        elif formatType.lower() != "lfn":
            raise NotImplementedError(formatType)
        yield self.threadTask(bk.writeJobOptions, exists, fileName, **kwargs)

        tarFile = "%s.tar.gz" % tmpdir
        with tarfile.open(tarFile, "w:gz") as tar:
            tar.add(tmpdir, arcname=userName)

        with open(tarFile, "rb") as tar:
            data = tar.read()
        shutil.rmtree(tmpdir)
        shutil.os.remove(tarFile)
        del os.environ["X509_USER_PROXY"]
        shutil.os.remove(proxyFile)

        self.set_header("Content-type", "application/gzip")
        self.set_header("Content-Disposition", 'attachment; filename="%s"' % tarFile)
        self.set_header("Content-Length", len(data))
        self.set_header("Cache-Control", "no-cache, no-store, must-revalidate, max-age=0")
        self.set_header("Pragma", "no-cache")
        self.finish(data)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import json

from DIRAC import gLogger
from DIRAC.Core.Utilities import Time

from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

from WebAppDIRAC.WebApp.handler.TransformationMonitorHandler import TransformationMonitorHandler
from WebAppDIRAC.Lib.WebHandler import asyncGen, WErr
from WebAppDIRAC.Lib.SessionData import SessionData


class LHCbTransformationMonitorHandler(TransformationMonitorHandler):

    AUTH_PROPS = "authenticated"

    def index(self):
        pass

    def web_standalone(self):
        self.render(
            "TransformationMonitorHandler/standalone.tpl", config_data=json.dumps(SessionData(None, None).getData())
        )

    ################################################################################
    def _TransformationMonitorHandler__dataQuery(self, prodid):
        callback = {}

        tsClient = TransformationClient()
        res = tsClient.getBookkeepingQuery(prodid)
        gLogger.info("-= #######", res)
        if not res["OK"]:
            callback = {"success": "false", "error": res["Message"]}
        else:
            result = res["Value"]
            back = []
            for i in sorted(result.keys(), reverse=False):
                back.append([i, result[i]])
            callback = {"success": "true", "result": back}
        return callback

    ################################################################################
    @asyncGen
    def web_showRunStatus(self):
        callback = {}
        start = int(self.get_argument("start"))
        limit = int(self.get_argument("limit"))

        try:
            id = int(self.get_argument("TransformationId"))
        except KeyError as excp:
            raise WErr(400, "Missing %s" % excp)

        tsClient = TransformationClient()
        result = yield self.threadTask(
            tsClient.getTransformationRunsSummaryWeb, {"TransformationID": id}, [["RunNumber", "DESC"]], start, limit
        )

        if not result["OK"]:
            callback = {"success": "false", "error": result["Message"]}
        else:
            result = result["Value"]
            if "TotalRecords" in result and result["TotalRecords"] > 0:
                total = result["TotalRecords"]
                if "Extras" in result:
                    extra = result["Extras"]
                if "ParameterNames" in result and "Records" in result:
                    head = result["ParameterNames"]
                    if len(head) > 0:
                        headLength = len(head)
                        if len(result["Records"]) > 0:
                            callback = []
                            jobs = result["Records"]
                            for i in jobs:
                                if len(i) != headLength:
                                    gLogger.info("Faulty record: %s" % i)
                                    callback = {
                                        "success": "false",
                                        "result": callback,
                                        "total": total,
                                        "error": "One of the records in service response is corrupted",
                                    }
                                    self.finish(callback)
                                tmp = {}
                                for j in range(0, headLength):
                                    tmp[head[j]] = i[j]
                                callback.append(tmp)
                            timestamp = Time.dateTime().strftime("%Y-%m-%d %H:%M [UTC]")
                            if extra:
                                callback = {
                                    "success": "true",
                                    "result": callback,
                                    "total": total,
                                    "extra": extra,
                                    "date": timestamp,
                                }
                            else:
                                callback = {"success": "true", "result": callback, "total": total, "date": timestamp}
                        else:
                            callback = {"success": "false", "result": "", "error": "There are no data to display"}
                    else:
                        callback = {"success": "false", "result": "", "error": "ParameterNames field is undefined"}
                else:
                    callback = {"success": "false", "result": "", "error": "Data structure is corrupted"}
            else:
                callback = {"success": "false", "result": "", "error": "There were no data matching your selection"}
        self.finish(callback)

    ################################################################################
    @asyncGen
    def web_setRunStatus(self):
        callback = {}
        transID = int(self.get_argument("TransformationId"))
        runID = int(self.get_argument("RunNumber"))
        status = self.get_argument("Status")

        gLogger.info("\033[0;31m setTransformationRunStatus(%s, %s, %s) \033[0m" % (transID, runID, status))
        tsClient = TransformationClient()
        result = result = yield self.threadTask(tsClient.setTransformationRunStatus, transID, runID, status)
        if result["OK"]:
            callback = {"success": True, "result": True}
        else:
            callback = {"success": "false", "error": result["Message"]}
        self.finish(callback)

    def _request(self):
        req = super(LHCbTransformationMonitorHandler, self)._request()
        hotFlag = json.loads(self.get_argument("Hot", "[false]"))[-1]
        if hotFlag:
            req["Hot"] = hotFlag

        return req

    @asyncGen
    def web_changeHotFlag(self):
        data = self.getSessionData()
        isAuth = False
        if "JobAdministrator" in data.get("user", {}).get("properties", {}):
            isAuth = True
        if not isAuth:
            raise WErr(500, "You are not authorized to change the hot flag (only lhcb_prmgr can change it)!!")

        hotFlag = json.loads(self.get_argument("Hot"))
        prod = int(self.get_argument("Production"))

        tsClient = TransformationClient()

        retVal = yield self.threadTask(tsClient.setHotFlag, prod, hotFlag)
        if not retVal["OK"]:
            raise WErr.fromSERROR(retVal)
        self.finish({"success": True, "result": prod})



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import datetime
import json
from datetime import timedelta
from time import time

from DIRAC import gLogger
from DIRAC.Core.Utilities import Time
from DIRAC.Core.Utilities.Graphs.Palette import Palette
from DIRAC.Core.DISET.RPCClient import RPCClient

from WebAppDIRAC.Lib.WebHandler import WebHandler, asyncGen


class RAWIntegrityMonitorHandler(WebHandler):

    AUTH_PROPS = "authenticated"

    numberOfFiles = 25
    pageNumber = 0
    globalSort = [["SubmitTime", "DESC"]]

    def index(self):
        pass

    @asyncGen
    def web_getSelectionData(self):
        callback = {}
        now = datetime.datetime.today()
        mon = timedelta(days=30)
        sd = now - mon
        tmp = {}
        ttt = str(sd.isoformat())
        gLogger.info(" - T I M E - ", ttt)
        tmp["startDate"] = sd.isoformat()
        tmp["startTime"] = sd.isoformat()
        callback["extra"] = tmp
        if len(self.request.arguments) > 0:
            tmp = {}
            for i in self.request.arguments:
                tmp[i] = self.get_arguments(i)
            callback["extra"] = tmp
        #####################################################################
        # This is the part that obtains the selections from the integrity db.
        RPC = RPCClient("DataManagement/RAWIntegrity")

        result = yield self.threadTask(RPC.getFileSelections)

        if result["OK"]:
            if len(result["Value"]) > 0:
                result = result["Value"]
                for key, value in result.items():
                    if len(value) > 3:
                        value = ["All"] + value
                    key = key.lower()
                    value = map(lambda x: [x], value)
                    callback[key] = value
        else:
            callback = {"success": "false", "error": result["Message"]}
        gLogger.info(" - callback - ", callback)
        self.finish(callback)

    def __request(self):
        req = {}
        lfns = list(json.loads(self.get_argument("lfn", "[]")))
        if lfns:
            req["lfn"] = lfns

        self.numberOfFiles = int(self.get_argument("limit", "25"))
        self.pageNumber = int(self.get_argument("start", "0"))

        #######################################################################
        # For the selection boxes only
        if "status" in self.request.arguments:
            req["Status"] = list(json.loads(self.get_argument("status")))

        if "storageelement" in self.request.arguments:
            if str(self.request.params["storageelement"]) != "All":
                req["StorageElement"] = list(json.loads(self.get_argument("storageelement")))
        #######################################################################
        # For the start time selection
        if "startDate" in self.request.arguments and len(self.get_argument("startDate")) > 0:
            if "startTime" in self.request.arguments and len(self.get_argument("startTime")) > 0:
                req["FromDate"] = str(self.get_argument("startDate") + " " + self.get_argument("startTime"))
            else:
                req["FromDate"] = self.get_argument("startDate")

        if "endDate" in self.request.arguments and len(self.get_argument("endDate")) > 0:
            if "endTime" in self.request.arguments and len(self.get_argument("endTime")) > 0:
                req["ToDate"] = str(self.get_argument("endDate") + " " + self.get_argument("endTime"))
            else:
                req["ToDate"] = self.get_argument("endDate")

        #######################################################################
        # The global sort of the data
        if "sort" in self.request.arguments:
            sort = json.loads(self.get_argument("sort"))
            if len(sort) > 0:
                self.globalSort = []
                for i in sort:
                    self.globalSort += [[i["property"], i["direction"]]]
        else:
            self.globalSort = [["SubmitTime", "DESC"]]

        gLogger.info("REQUEST:", req)
        return req

    @asyncGen
    def web_getloggingInfo(self):
        req = self.__request()
        gLogger.debug("getloggingInfo" + str(req))

        RPC = RPCClient("DataManagement/DataLogging")
        result = yield self.threadTask(RPC.getFileLoggingInfo, str(req["lfn"][0]))

        if not result["OK"]:
            self.finish({"success": "false", "error": result["Message"]})
        result = result["Value"]
        if not result:
            self.finish({"success": "false", "result": "", "error": "No logging information found for LFN"})
        callback = []
        for i in result:
            callback.append([i[0], i[1], i[2], i[3]])
        self.finish({"success": "true", "result": callback})

    @asyncGen
    def web_getStatisticsData(self):
        paletteColor = Palette()
        gLogger.debug("Params:" + str(self.request.arguments))

        req = self.__request()
        selector = self.get_argument("statsField")

        if selector == "Status":
            selector = "status"
        if selector == "Storage Element":
            selector = "storageelement"

        RPC = RPCClient("DataManagement/RAWIntegrity")

        result = yield self.threadTask(RPC.getStatistics, selector, req)
        gLogger.info(" - result - :", result)
        callback = {}

        if result["OK"]:
            callback = []
            result = result["Value"]
            keylist = sorted(result.keys())

            for key in keylist:
                callback.append({"key": key, "value": result[key], "code": "", "color": paletteColor.getColor(key)})
            callback = {"success": "true", "result": callback}
        else:
            callback = {"success": "false", "error": result["Message"]}

        gLogger.debug("retValue" + str(callback))
        self.finish(callback)

    #####################################################################
    #
    # Handles displaying results
    #
    @asyncGen
    def web_getRawIntegrityData(self):
        gLogger.info(" -- SUBMIT --")
        pagestart = time()
        RPC = RPCClient("DataManagement/RAWIntegrity")
        result = self.__request()
        result = yield self.threadTask(
            RPC.getFilesSummaryWeb, result, self.globalSort, self.pageNumber, self.numberOfFiles
        )
        if result["OK"]:
            result = result["Value"]

            if "TotalRecords" in result:
                if result["TotalRecords"] > 0:
                    if "ParameterNames" in result and "Records" in result:
                        if len(result["ParameterNames"]) > 0:
                            if len(result["Records"]) > 0:
                                callback = []
                                jobs = result["Records"]
                                head = result["ParameterNames"]
                                headLength = len(head)
                                for i in jobs:
                                    tmp = {}
                                    for j in range(0, headLength):
                                        tmp[head[j]] = i[j]
                                    callback.append(tmp)
                                total = result["TotalRecords"]
                                if "Extras" in result:
                                    extra = result["Extras"]
                                    callback = {"success": "true", "result": callback, "total": total, "extra": extra}
                                else:
                                    callback = {"success": "true", "result": callback, "total": total}
                                timestamp = Time.dateTime().strftime("%Y-%m-%d %H:%M [UTC]")
                                callback["date"] = timestamp
                            else:
                                callback = {"success": "false", "result": "", "error": "There are no data to display"}
                        else:
                            callback = {"success": "false", "result": "", "error": "ParameterNames field is missing"}
                    else:
                        callback = {"success": "false", "result": "", "error": "Data structure is corrupted"}
                else:
                    callback = {"success": "false", "result": "", "error": "There were no data matching your selection"}
            else:
                callback = {"success": "false", "result": "", "error": "Data structure is corrupted"}
        else:
            callback = {"success": "false", "error": result["Message"]}
        gLogger.info("\033[0;31mJOB SUBMIT REQUEST:\033[0m %s" % (time() - pagestart))
        self.finish(callback)
